"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import os
import gemmi
import numpy
import pandas
import scipy.sparse
import servalcat # for version
from servalcat.utils import logger
from servalcat import utils
from servalcat.refmac import exte
from servalcat.refmac.refmac_keywords import parse_keywords
from . import cgsolve
u_to_b = utils.model.u_to_b
b_to_u = utils.model.b_to_u

#import line_profiler
#import atexit
#profile = line_profiler.LineProfiler()
#atexit.register(profile.print_stats)

class Geom:
    def __init__(self, st, topo, monlib, sigma_b=30, shake_rms=0, refmac_keywords=None):
        self.st = st
        self.lookup = {x.atom: x for x in self.st[0].all()}
        self.specs = utils.model.find_special_positions(self.st)
        self.geom = gemmi.Geometry(self.st, monlib.ener_lib)
        self.sigma_b = sigma_b
        if shake_rms > 0:
            numpy.random.seed(0)
            utils.model.shake_structure(self.st, shake_rms, copy=False)
            utils.fileio.write_model(self.st, "shaken", pdb=True, cif=True)
        self.geom.load_topo(topo)
        self.use_nucleus = False
        self.calc_kwds = {"use_nucleus": self.use_nucleus}
        if refmac_keywords:
            exte.read_external_restraints(refmac_keywords, self.st, self.geom)
            kwds = parse_keywords(refmac_keywords)
            for k in ("wbond", "wangle", "wtors", "wplane", "wchir", "wvdw"):
                if k in kwds:
                    self.calc_kwds[k] = kwds[k]
                    logger.writeln("setting geometry weight {}= {}".format(k, kwds[k]))
        self.geom.finalize_restraints()
        self.outlier_sigmas = dict(bond=5, angle=5, torsion=5, vdw=5, chir=5, plane=5, staca=5, stacd=5)
        self.parents = {}
        self.check_chemtypes(os.path.join(monlib.path(), "ener_lib.cif"), topo)
    # __init__()

    def check_chemtypes(self, enerlib_path, topo):
        block = gemmi.cif.read(enerlib_path).sole_block()
        all_types = set(block.find_values("_lib_atom.type"))
        for ci in topo.chain_infos:
            for ri in ci.res_infos:
                cc_all = {x: ri.get_final_chemcomp(x) for x in set(a.altloc for a in ri.res)}
                for a in ri.res:
                    cca = cc_all[a.altloc].find_atom(a.name)
                    if cca is None: # I believe it won't happen..
                        logger.writeln("WARNING: restraint for {} not found.".format(self.lookup[a]))
                    elif cca.chem_type not in all_types:
                        raise RuntimeError("Energy type {} of {} not found in ener_lib.".format(cca.chem_type,
                                                                                                self.lookup[a]))
    def set_h_parents(self):
        self.parents = {}
        for bond in self.geom.bonds:
            if bond.atoms[0].is_hydrogen():
                self.parents[bond.atoms[0]] = bond.atoms[1]
            elif bond.atoms[1].is_hydrogen():
                self.parents[bond.atoms[1]] = bond.atoms[0]
    # set_h_parents()

    def calc(self, target_only):
        return self.geom.calc(check_only=target_only, **self.calc_kwds)
    def calc_adp_restraint(self, target_only):
        return self.geom.calc_adp_restraint(target_only, self.sigma_b)
    def calc_target(self, target_only, refine_xyz, adp_mode, N):
        self.geom.clear_target()
        geom_x = self.calc(target_only) if refine_xyz else 0
        geom_a = self.calc_adp_restraint(target_only) if adp_mode > 0 else 0
        logger.writeln(" geom_x = {}".format(geom_x))
        logger.writeln(" geom_a = {}".format(geom_a))
        geom = geom_x + geom_a
        if not target_only:
            g_vn = numpy.array(self.geom.target.vn) # don't want copy?            
            coo = scipy.sparse.coo_matrix(self.geom.target.am_for_coo(), shape=(N, N))
            lil = coo.tolil()
            rows, cols = lil.nonzero()
            lil[cols,rows] = lil[rows,cols]
            g_am = lil
            diag = g_am.diagonal()
            logger.writeln("diag(restr) min= {:3e} max= {:3e}".format(numpy.min(diag),
                                                                      numpy.max(diag)))
        else:
            g_vn, g_am = None, None

        return geom, g_vn, g_am
        
    def show_model_stats(self, show_outliers=True):
        f0 = self.calc(True)
        ret = {"outliers": {}}
        if show_outliers:
            get_table = dict(bond=self.geom.reporting.get_bond_outliers,
                             angle=self.geom.reporting.get_angle_outliers,
                             torsion=self.geom.reporting.get_torsion_outliers,
                             chir=self.geom.reporting.get_chiral_outliers,
                             plane=self.geom.reporting.get_plane_outliers,
                             staca=self.geom.reporting.get_stacking_angle_outliers,
                             stacd=self.geom.reporting.get_stacking_dist_outliers,
                             vdw=self.geom.reporting.get_vdw_outliers,
                             )
            labs = dict(bond="Bond distances",
                        angle="Bond angles",
                        torsion="Torsion angles",
                        chir="Chiral centres",
                        plane="Planar groups",
                        staca="Stacking plane angles",
                        stacd="Stacking plane distances",
                        vdw="VDW repulsions")

            for k in get_table:
                kwgs = {"min_z": self.outlier_sigmas[k]}
                if k == "bond": kwgs["use_nucleus"] = self.use_nucleus
                table = get_table[k](**kwgs)
                if table["z"]:
                    for kk in table:
                        if kk.startswith(("atom", "plane")):
                            table[kk] = [str(self.lookup[x]) for x in table[kk]]
                    df = pandas.DataFrame(table)
                    df = df.reindex(df.z.abs().sort_values(ascending=False).index)
                    ret["outliers"][k] = df
                    if k == "bond":
                        df0 = df[df.type < 2].drop(columns=["type", "alpha"])
                        if len(df0.index) > 0:
                            logger.writeln(" *** {} outliers (Z >= {}) ***\n".format(labs[k], self.outlier_sigmas[k]))
                            logger.writeln(df0.to_string(float_format="{:.3f}".format, index=False) + "\n")
                        df0 = df[df.type == 2].drop(columns=["type"])
                        if len(df0.index) > 0:
                            logger.writeln(" *** External bond outliers (Z >= {}) ***\n".format(self.outlier_sigmas[k]))
                            logger.writeln(df0.to_string(float_format="{:.3f}".format, index=False) + "\n")
                    else:
                        logger.writeln(" *** {} outliers (Z >= {}) ***\n".format(labs[k], self.outlier_sigmas[k]))
                        logger.writeln(df.to_string(float_format="{:.3f}".format, index=False) + "\n")

        df = pandas.DataFrame(self.geom.reporting.get_summary_table(self.use_nucleus))
        df = df.set_index("Restraint type").rename_axis(index=None)
        ret["summary"] = df
        logger.writeln(df.to_string(float_format="{:.3f}".format) + "\n")
        return ret
        
class Refine:
    def __init__(self, st, geom, ll=None, refine_xyz=True, adp_mode=1, refine_h=False, unrestrained=False):
        assert adp_mode in (0, 1, 2) # 0=fix, 1=iso, 2=aniso
        assert geom is not None
        self.st = st # clone()?
        self.atoms = [None for _ in range(self.st[0].count_atom_sites())]
        for cra in self.st[0].all(): self.atoms[cra.atom.serial-1] = cra.atom
        self.geom = geom
        self.ll = ll
        self.gamma = 0
        self.adp_mode = 0 if self.ll is None else adp_mode
        self.refine_xyz = refine_xyz
        self.unrestrained = unrestrained
        self.refine_h = refine_h
        self.h_inherit_parent_adp = self.adp_mode > 0 and not self.refine_h and self.st[0].has_hydrogen()
        if self.h_inherit_parent_adp:
            self.geom.set_h_parents()
    # __init__()

    def scale_shifts(self, dx, scale):
        n_atoms = len(self.atoms)
        #ave_shift = numpy.mean(dx)
        #max_shift = numpy.maximum(dx)
        #rms_shift = numpy.std(dx)
        shift_allow_high =  0.5
        shift_allow_low  = -0.5
        shift_max_allow_B = 30.0
        shift_min_allow_B = -30.0
        dx = scale * dx
        offset_b = 0
        if self.refine_xyz:
            dxx = dx[:n_atoms*3]
            dxx[dxx > shift_allow_high] = shift_allow_high
            dxx[dxx < shift_allow_low] = shift_allow_low
            offset_b = n_atoms*3
        if self.adp_mode == 1:
            dxb = dx[offset_b:]
            dxb[dxb > shift_max_allow_B] = shift_max_allow_B
            dxb[dxb < shift_min_allow_B] = shift_min_allow_B
        elif self.adp_mode == 2: # FIXME aniso. need to find eigenvalues..
            dxb = dx[offset_b:]
            dxb[dxb > shift_max_allow_B] = shift_max_allow_B
            dxb[dxb < shift_min_allow_B] = shift_min_allow_B
            
        return dx

    def n_params(self):
        n_atoms = len(self.atoms)
        n_params = 0
        if self.refine_xyz: n_params += 3 * n_atoms
        if self.adp_mode == 1:
            n_params += n_atoms
        elif self.adp_mode == 2:
            n_params += 6 * n_atoms
        return n_params

    def set_x(self, x):
        n_atoms = len(self.atoms)
        offset_b = n_atoms * 3 if self.refine_xyz else 0
        for i in range(len(self.atoms)):
            if self.refine_xyz:
                self.atoms[i].pos.fromlist(x[3*i:3*i+3]) # faster than substituting pos.x,pos.y,pos.z
            if self.adp_mode == 1:
                self.atoms[i].b_iso = max(0.5, x[offset_b + i]) # minimum B = 0.5
            elif self.adp_mode == 2:
                a = x[offset_b + 6 * i: offset_b + 6 * (i+1)]
                a = gemmi.SMat33d(*a)
                M = numpy.array(a.as_mat33())
                v, Q = numpy.linalg.eigh(M) # eig() may return complex due to numerical precision?
                v = numpy.maximum(v, 0.5) # avoid NPD with minimum B = 0.5
                M2 = Q.dot(numpy.diag(v)).dot(Q.T)
                self.atoms[i].b_iso = M2.trace() / 3
                M2 *= b_to_u
                self.atoms[i].aniso = gemmi.SMat33f(M2[0,0], M2[1,1], M2[2,2], M2[0,1], M2[0,2], M2[1,2])

        # Copy B of hydrogen from parent
        if self.h_inherit_parent_adp:
            for h in self.geom.parents:
                p = self.geom.parents[h]
                h.b_iso = p.b_iso
                h.aniso = p.aniso

        if self.ll is not None:
            self.ll.update_fc()
        
        self.geom.geom.setup_nonbonded() # if refine_xyz=False, no need to do it every time
        logger.writeln("vdws = {}".format(len(self.geom.geom.vdws)))

    def get_x(self):
        n_atoms = len(self.atoms)
        offset_b = n_atoms * 3 if self.refine_xyz else 0
        x = numpy.zeros(self.n_params())
        for i, a in enumerate(self.atoms):
            if self.refine_xyz:
                x[3*i:3*(i+1)] = a.pos.tolist()
            if self.adp_mode == 1:
                x[offset_b + i] = self.atoms[i].b_iso
            elif self.adp_mode == 2:
                x[offset_b + 6*i : offset_b + 6*(i+1)] = self.atoms[i].aniso.elements_pdb()
                x[offset_b + 6*i : offset_b + 6*(i+1)] *= u_to_b

        return x
    #@profile
    def calc_target(self, w=1, target_only=False):
        N = self.n_params()
        geom, g_vn, g_am = self.geom.calc_target(target_only,
                                                 not self.unrestrained and self.refine_xyz,
                                                 self.adp_mode, N)
        if self.ll is not None:
            ll = self.ll.calc_target()
            if not target_only:
                l_vn, l_am = self.ll.calc_grad(self.refine_xyz, self.adp_mode, self.refine_h, self.geom.specs)
                diag = l_am.diagonal()
                logger.writeln("diag(data) min= {:3e} max= {:3e}".format(numpy.min(diag),
                                                                         numpy.max(diag)))
        else:
            ll = 0

        f =  w * ll + geom

        if not target_only:
            if self.ll is not None:
                vn = w * l_vn + g_vn
                am = w * l_am + g_am
            else:
                vn = g_vn
                am = g_am
        else:
            vn, am = None, None

        return f, vn, am

    #@profile
    def run_cycle(self, weight=1):
        if self.ll is not None:
            self.ll.overall_scale()
            self.ll.update_ml_params()

        self.geom.geom.setup_target(self.refine_xyz, self.adp_mode)
            
        if 0: # test of grad
            self.ll.update_fc()
            x0 = self.get_x()
            f0,ader,_ = self.calc_target(weight)
            i = 1
            for e in 1e-1,1e-2,1e-3, 1e-4, 1e-5:
                x1 = numpy.copy(x0)
                x1[i] += e
                self.set_x(x1)
                self.ll.update_fc()
                f1,_,_ = self.calc_target(weight, target_only=True)
                nder = (f1 - f0) / e
                print("e=", e)
                print("NUM DER=", nder)
                print("ANA DER=", ader[i])
                print("ratio=", nder/ader[i])
            quit()

        f0, vn, am = self.calc_target(weight)
        x0 = self.get_x()
        logger.writeln("f0= {:.4e}".format(f0))
        
        if 0:
            assert self.adp_mode == 0 # not supported!
            logger.writeln("Preconditioning using eigen")
            #Pinv = scipy.sparse.coo_matrix(self.geom.target.precondition_eigen_coo(1e-4), shape=(N, N)) # did not work if <= 1e-7
            N = len(self.atoms) * 3
            tmp = am.tocoo()
            Pinv = scipy.sparse.coo_matrix(gemmi.precondition_eigen_coo(tmp.data, tmp.row, tmp.col, N, 1e-4),
                                           shape=(N, N))
            M = Pinv
            #a = Pinv.T.dot(lil).dot(Pinv)
            #logger.writeln("cond(Pinv^-1 A Pinv)= {}".format(numpy.linalg.cond(a.todense())))
            #logger.writeln("cond(A)= {}".format(numpy.linalg.cond(lil.todense())))
            #quit()
            #M = scipy.sparse.identity(N)
        else:
            diag = am.diagonal()
            logger.writeln("diagonal min= {:3e} max= {:3e}".format(numpy.min(diag),
                                                                   numpy.max(diag)))
            #for i in numpy.where(diag <= 0)[0]:
            #    if self.refine_xyz:
            #        if i >= len(self.atoms)*3:
            #            j = i - len(self.atoms)*3
            #            print("diag<=0: {} B {}".format(i, self.geom.lookup[self.atoms[j]]))
            #        else:
            #            j = i // 3
            #            print("diag<=0: {} x {}".format(i, self.geom.lookup[self.atoms[j]]))
            diag[diag<=0] = 1.
            diag = numpy.sqrt(diag)
            rdiag = 1./diag # sk
            rdmat = scipy.sparse.diags(rdiag)
            M = rdmat
            
        dx, self.gamma = cgsolve.cgsolve_rm(A=am, v=vn, M=M, gamma=self.gamma)
        if self.refine_xyz:
            dxx = dx[:len(self.atoms)*3]
            #logger.writeln("dx = {}".format(dxx))
            logger.writeln("min(dx) = {}".format(numpy.min(dxx)))
            logger.writeln("max(dx) = {}".format(numpy.max(dxx)))
            logger.writeln("mean(dx)= {}".format(numpy.mean(dxx)))
        if self.adp_mode > 0: # TODO for aniso
            db = dx[len(self.atoms)*3 if self.refine_xyz else 0:]
            #logger.writeln("dB = {}".format(db))
            logger.writeln("min(dB) = {}".format(numpy.min(db)))
            logger.writeln("max(dB) = {}".format(numpy.max(db)))
            logger.writeln("mean(dB)= {}".format(numpy.mean(db)))
            
        if 0: # to check hessian scale
            with open("minimise_line.dat", "w") as ofs:
                ofs.write("s f\n")
                for s in numpy.arange(-2, 2, 0.1):
                    self.set_x(x0 + s * dx)
                    fval = self.calc_target(weight, target_only=True)[0]
                    ofs.write("{} {}\n".format(s, fval))
            quit()

        ret = True # success
        for i in range(3):
            dx2 = self.scale_shifts(dx, 1/2**i)
            self.set_x(x0 - dx2)
            f1, _, _ = self.calc_target(weight, target_only=True)
            logger.writeln("f1, {}= {:.4e}".format(i, f1))
            if f1 < f0: break
        else:
            ret = False
            logger.writeln("WARNING: function not minimised")
            #self.set_x(x0) # Refmac accepts it even when function increases

        return ret
    
    def run_cycles(self, ncycles, weight=1, debug=False):
        stats = [{"Ncyc": 0}]
        self.geom.geom.setup_nonbonded()
        logger.writeln("vdws = {}".format(len(self.geom.geom.vdws)))
        if self.refine_xyz and not self.unrestrained:
            stats[-1]["geom"] = self.geom.show_model_stats(show_outliers=True)["summary"]
        if self.ll is not None:
            self.ll.update_fc()
            self.ll.overall_scale()
            self.ll.update_ml_params()
            stats[-1]["data"] = self.ll.calc_stats()["summary"]
            
        for i in range(ncycles):
            logger.writeln("\n====== CYCLE {:2d} ======\n".format(i+1))
            self.run_cycle(weight=weight) # check ret?
            stats.append({"Ncyc": i+1})
            if debug: utils.fileio.write_model(self.st, "refined_{:02d}".format(i+1), pdb=True)#, cif=True)
            if self.refine_xyz and not self.unrestrained:
                stats[-1]["geom"] = self.geom.show_model_stats(show_outliers=(i==ncycles-1))["summary"]
            if self.adp_mode > 0:
                utils.model.adp_analysis(self.st)
            logger.writeln("")

            if self.ll is not None:
                self.ll.overall_scale()
                stats[-1]["data"] = self.ll.calc_stats()["summary"]

        df = pandas.DataFrame({"Ncyc": range(ncycles+1)})
        if self.ll is not None:
            df["FSCaverage"] = [s["data"].get("FSCaverage", numpy.nan) for s in stats]
            for r in ("R", "Rwork", "Rfree"):
                df[r] = [s["data"].get(r, numpy.nan) for s in stats]
            df["-LL"] = [s["data"].get("-LL", numpy.nan) for s in stats]
        if self.refine_xyz and not self.unrestrained:
            df["rmsBOND"] =[s["geom"]["r.m.s.d."].get("Bond distances, non H") for s in stats]
            df["zBOND"] = [s["geom"]["r.m.s.Z"].get("Bond distances, non H") for s in stats]
            df["rmsANGL"] = [s["geom"]["r.m.s.d."].get("Bond angles, non H") for s in stats]
            df["zANGL"] = [s["geom"]["r.m.s.Z"].get("Bond angles, non H") for s in stats]

        forplot = []
        if self.ll is not None:
            forplot.append(["FSC or R", ["Ncyc", "FSCaverage", "R"]])
            forplot.append(["-LL", ["Ncyc", "-LL"]])
        if self.refine_xyz and not self.unrestrained:
            forplot.extend([["Geometry", ["Ncyc", "rmsBOND", "rmsANGL"]],
                            ["Geometry Z", ["Ncyc", "zBOND", "zANGL"]]])

        lstr = utils.make_loggraph_str(df, "stats vs cycle", forplot,
                                       float_format="{:.4f}".format)
        logger.writeln(lstr)
        self.update_meta()
        return stats

    def update_meta(self):
        # TODO write stats. probably geom.reporting.get_summary_table should return with _refine_ls_restr.type names
        self.st.raw_remarks = []
        si = gemmi.SoftwareItem()
        si.classification = gemmi.SoftwareItem.Classification.Refinement
        si.pdbx_ordinal = 1
        si.name = "Servalcat"
        si.version = servalcat.__version__
        si.date = servalcat.__date__
        self.st.meta.software = [si]

        self.st.meta.refinement = []
        #ri = gemmi.RefinementInfo()
        #rr = gemmi.RefinementInfo.Restr("")
        #ri.restr_stats.append(rr)
        #st.meta.refinement = [ri]
        
# class Refine



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import json
import scipy.sparse
from servalcat.utils import logger
from servalcat import utils
from servalcat.spa import fofc
from servalcat.spa import fsc
b_to_u = utils.model.b_to_u
u_to_b = utils.model.u_to_b

def calc_D_and_S(hkldata, lab_obs): # simplified version of fofc.calc_D_and_S()
    bdf = hkldata.binned_df
    bdf["D"] = 0.
    bdf["S"] = 0.
    for i_bin, idxes in hkldata.binned():
        Fo = hkldata.df[lab_obs].to_numpy()[idxes]
        Fc = hkldata.df.FC.to_numpy()[idxes]
        bdf.loc[i_bin, "D"] = numpy.nansum(numpy.real(Fo * numpy.conj(Fc))) / numpy.sum(numpy.abs(Fc)**2)
        bdf.loc[i_bin, "S"] = numpy.nanmean(numpy.abs(Fo - bdf.D[i_bin] * Fc)**2)
# calc_D_and_S()

class LL_SPA:
    def __init__(self, hkldata, st, monlib, lab_obs, source="electron", mott_bethe=True):
        assert source in ("electron", "xray")
        self.source = source
        self.mott_bethe = False if source != "electron" else mott_bethe
        self.hkldata = hkldata
        self.lab_obs = lab_obs
        self.st = st
        self.monlib = monlib
        self.d_min = hkldata.d_min_max()[0]

    def update_ml_params(self):
        # FIXME make sure D > 0
        calc_D_and_S(self.hkldata, self.lab_obs)
        logger.writeln(self.hkldata.binned_df.to_string(columns=["d_max", "d_min", "D", "S"]))

    def update_fc(self):
        if self.st.ncs:
            st = self.st.clone()
            st.expand_ncs(gemmi.HowToNameCopiedChain.Dup)
        else:
            st = self.st

        self.hkldata.df["FC"] = utils.model.calc_fc_fft(st, self.d_min - 1e-6,
                                                        monlib=self.monlib,
                                                        source=self.source,
                                                        mott_bethe=self.mott_bethe,
                                                        miller_array=self.hkldata.miller_array())

    def overall_scale(self, min_b=0.5):
        k, b = self.hkldata.scale_k_and_b(lab_ref=self.lab_obs, lab_scaled="FC")
        min_b_iso = utils.model.minimum_b(self.st[0]) # actually min of aniso too
        tmp = min_b_iso + b
        if tmp < min_b: # perhaps better only adjust b_iso that went too small, but we need to recalculate Fc
            logger.writeln("Adjusting overall B to avoid too small value")
            b += min_b - tmp
        logger.writeln("Applying overall B to model: {:.2f}".format(b))
        utils.model.shift_b(self.st[0], b)
        # adjust Fc
        k_iso = self.hkldata.debye_waller_factors(b_iso=b)
        self.hkldata.df["FC"] *= k_iso
    # overall_scale()

    def calc_target(self): # -LL target for SPA
        ret = 0
        for i_bin, idxes in self.hkldata.binned():
            Fo = self.hkldata.df[self.lab_obs].to_numpy()[idxes]
            DFc = self.hkldata.df.FC.to_numpy()[idxes] * self.hkldata.binned_df.D[i_bin]
            ret += numpy.nansum(numpy.abs(Fo - DFc)**2) / self.hkldata.binned_df.S[i_bin]
        return ret * 2 # friedel mates
    # calc_target()

    def calc_stats(self):
        stats = fsc.calc_fsc_all(self.hkldata, labs_fc=["FC"], lab_f=self.lab_obs)
        fsca = fsc.fsc_average(stats.ncoeffs, stats.fsc_FC_full)
        logger.writeln("FSCaverage = {:.4f}".format(fsca))
        return {"fsc": stats, "summary": {"FSCaverage": fsca, "-LL": self.calc_target()}}

    def calc_grad(self, refine_xyz, adp_mode, refine_h, specs): # specs not used
        dll_dab = numpy.empty_like(self.hkldata.df[self.lab_obs])
        d2ll_dab2 = numpy.zeros(len(self.hkldata.df.index))
        blur = utils.model.determine_blur_for_dencalc(self.st, self.d_min / 3) # TODO need more work
        logger.writeln("blur for deriv= {:.2f}".format(blur))
        for i_bin, idxes in self.hkldata.binned():
            D = self.hkldata.binned_df.D[i_bin]
            S = self.hkldata.binned_df.S[i_bin]
            Fc = self.hkldata.df.FC.to_numpy()[idxes]
            Fo = self.hkldata.df[self.lab_obs].to_numpy()[idxes]
            dll_dab[idxes] = -2 * D / S * (Fo - D * Fc)#.conj()
            d2ll_dab2[idxes] = 2 * D**2 / S

        if self.mott_bethe:
            dll_dab *= self.hkldata.d_spacings()**2 * gemmi.mott_bethe_const()
            d2ll_dab2 *= gemmi.mott_bethe_const()**2

        # strangely, we need V for Hessian and V**2/n for gradient.
        d2ll_dab2 *= self.hkldata.cell.volume
        dll_dab_den = self.hkldata.fft_map(data=dll_dab * self.hkldata.debye_waller_factors(b_iso=-blur))
        dll_dab_den.array[:] *= self.hkldata.cell.volume**2 / dll_dab_den.point_count
        #asu = dll_dab_den.masked_asu()
        #dll_dab_den.array[:] *= 1 - asu.mask_array # 0 to use
        #atoms = [x.atom for x in self.st[0].all()]
        atoms = [None for _ in range(self.st[0].count_atom_sites())]
        for cra in self.st[0].all(): atoms[cra.atom.serial-1] = cra.atom
        ll = gemmi.LLX(self.st.cell, self.hkldata.sg, atoms, self.mott_bethe, refine_xyz, adp_mode, refine_h)
        ll.set_ncs([x.tr for x in self.st.ncs if not x.given])
        vn = ll.calc_grad(dll_dab_den, blur)
        d2dfw_table = gemmi.TableS3(*self.hkldata.d_min_max())
        d2dfw_table.make_table(1./self.hkldata.d_spacings(), d2ll_dab2)

        b_iso_all = [cra.atom.aniso.trace() / 3 * u_to_b if cra.atom.aniso.nonzero() else cra.atom.b_iso
                     for cra in self.st[0].all()]
        b_iso_min = min(b_iso_all)
        b_iso_max = max(b_iso_all)
        elems = set(cra.atom.element for cra in self.st[0].all())
        b_sf_min = 0 #min(min(e.it92.b) for e in elems) # because there is constants
        b_sf_max = max(max(e.it92.b) for e in elems)
        fisher_b_min = b_iso_min + b_sf_min
        fisher_b_max = 2 * (b_iso_max + b_sf_max)
        logger.writeln("preparing fast Fisher table for B= {:.2f} - {:.2f}".format(fisher_b_min, fisher_b_max))
        ll.make_fisher_table_diag_fast(fisher_b_min, fisher_b_max, d2dfw_table)
        #json.dump(dict(b=ll.table_bs, pp1=ll.pp1, bb=ll.bb),
        #          open("ll_fisher.json", "w"), indent=True)
        #a, (b,c) = ll.fisher_for_coo()
        #json.dump(([float(x) for x in a], ([int(x) for x in b], [int(x) for x in c])), open("fisher.json", "w"))
        coo = scipy.sparse.coo_matrix(ll.fisher_for_coo())
        lil = coo.tolil()
        rows, cols = lil.nonzero()
        lil[cols,rows] = lil[rows,cols]
        return numpy.array(vn), lil



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import subprocess
import pipes
import json
import copy
import re
import os
import string
import itertools
import tempfile
from servalcat.utils import logger
from servalcat.utils import fileio

re_version = re.compile("#.* Refmac *version ([^ ]+) ")
re_error = re.compile('(warn|error *[:]|error *==|^error)', re.IGNORECASE)
re_outlier_start = re.compile("\*\*\*\*.*outliers")

def check_version(exe="refmac5"):
    p = subprocess.Popen([exe], shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         universal_newlines=True)
    p.stdin.write("end\n")
    p.stdin.close()
    ver = ()
    for l in iter(p.stdout.readline, ""):
        r_ver = re_version.search(l)
        if r_ver:
            logger.writeln("Refmac version: {}".format(r_ver.group(1)))
            ver = tuple(map(int, r_ver.group(1).split(".")))
    p.wait()
    return ver
# check_version()

def ensure_ccp4scr():
    tmpdir = os.environ.get("CCP4_SCR")
    if tmpdir:
        if os.path.isdir(tmpdir): # TODO check writability
            try:
                t = tempfile.TemporaryFile(dir=tmpdir)
                t.close()
                return
            except OSError:
                logger.writeln("Warning: cannot write files in CCP4_SCR= {}".format(tmpdir))
        else:
            try:
                os.makedirs(tmpdir)
                return
            except:
                logger.writeln("Warning: cannot create CCP4_SCR= {}".format(tmpdir))

    os.environ["CCP4_SCR"] = tempfile.mkdtemp(prefix="ccp4tmp")
    logger.writeln("Updated CCP4_SCR= {}".format(os.environ["CCP4_SCR"]))
# ensure_ccp4scr()

def external_restraints_json_to_keywords(json_in):
    ret = []
    with open(json_in) as f: exte_list = json.load(f)
    for e in exte_list:
        if "use" in e:
            ret.append("EXTERNAL USE {}".format(e["use"]))
        if "dmax" in e:
            ret.append("EXTERNAL DMAX {0}".format(e["dmax"]))
        if "weight_scale" in e:
            ret.append("EXTERNAL WEIGHT SCALE {0}".format(e["weight_scale"]))
        if "weight_gmwt" in e:
            ret.append("EXTERNAL WEIGHT GMWT {0}".format(e["weight_gmwt"]))
        if "file" in e:
            ret.append("@"+e["file"])

    return "\n".join(ret) + "\n"
# external_restraints_json_to_keywords()

def read_tls_file(tlsin):
    # TODO sort out L/S units - currently use Refmac tlsin/out as is
    # TODO change to gemmi::TlsGroup?
    
    groups = []
    with open(tlsin) as ifs:
        for l in ifs:
            l = l.strip()
            if l.startswith("TLS"):
                title = l[4:]
                groups.append(dict(title=title, ranges=[], origin=None, T=None, L=None, S=None))
            elif l.startswith("RANG"):
                r = l[l.index(" "):].strip()
                groups[-1]["ranges"].append(r)
            elif l.startswith("ORIG"):
                try:
                    groups[-1]["origin"] = gemmi.Position(*(float(x) for x in l.split()[1:]))
                except:
                    raise ValueError("Prolem with TLS file: {}".format(l))
            elif l.startswith("T   "):
                try:
                    groups[-1]["T"] = [float(x) for x in l.split()[1:7]]
                except:
                    raise ValueError("Prolem with TLS file: {}".format(l))
            elif l.startswith("L   "):
                try:
                    groups[-1]["L"] = [float(x) for x in l.split()[1:7]]
                except:
                    raise ValueError("Prolem with TLS file: {}".format(l))
            elif l.startswith("S   "):
                try:
                    groups[-1]["S"] = [float(x) for x in l.split()[1:10]]
                except:
                    raise ValueError("Prolem with TLS file: {}".format(l))

    return groups
# read_tls_file()

def write_tls_file(groups, tlsout):
    with open(tlsout, "w") as f:
        for g in groups:
            f.write("TLS {}\n".format(g["title"]))
            for r in g["ranges"]:
                f.write("RANGE {}\n".format(r))
            if g["origin"] is not None:
                f.write("ORIGIN ")
                f.write(" ".join("{:8.4f}".format(x) for x in g["origin"].tolist()))
                f.write("\n")
            for k in "TLS":
                if g[k] is not None:
                    f.write("{:4s}".format(k))
                    f.write(" ".join("{:8.4f}".format(x) for x in g[k]))
                    f.write("\n")
# write_tls_file()

class FixForRefmac:
    """
    Workaround for Refmac limitations
    - microheterogeneity
    - residue number > 9999

    XXX fix external restraints accordingly
    TODO fix _struct_conf, _struct_sheet_range, _pdbx_struct_sheet_hbond
    """
    def __init__(self, st, topo, fix_microheterogeneity=True, fix_resimax=True, fix_nonpolymer=True, add_gaps=False):
        self.MAXNUM = 9999
        self.fixes = []
        self.chainids = set(chain.name for chain in st[0])
        if fix_microheterogeneity:
            self.fix_microheterogeneity(st, topo)
        if add_gaps:
            self.add_gaps(st, topo)
        if fix_resimax: # This modifies chains, so topo will be broken
            self.fix_too_large_seqnum(st, topo)
        if fix_nonpolymer: # This modifies chains, so topo will be broken
            self.fix_nonpolymer(st)

    def new_chain_id(self, original_chain_id):
        # decide new chain ID
        for i in itertools.count(start=1):
            new_id = "{}{}".format(original_chain_id, i)
            if new_id not in self.chainids:
                self.chainids.add(new_id)
                return new_id

    def fix_metadata(self, st, changedict):
        # fix connections
        # changedict = dict(changes)
        aa2tuple = lambda aa: (aa.chain_name, aa.res_id.seqid.num, chr(ord(aa.res_id.seqid.icode)|0x20))
        for con in st.connections:
            for aa in (con.partner1, con.partner2):
                changeto = changedict.get(aa2tuple(aa))
                if changeto is not None:
                    aa.chain_name = changeto[0]
                    aa.res_id.seqid.num = changeto[1]
                    aa.res_id.seqid.icode = changeto[2]

    def add_gaps(self, st, topo):
        # Refmac (as of 5.8.0352) has a bug that makes two links for IAS (IAS-pept and usual TRANS/CIS)
        # However this implementation is even more harmful.. if gap is inserted to real gaps then necessary p link is also gone!
        for chain in st[0]:
            rs = chain.get_polymer()
            for i in range(1, len(rs)):
                res0 = rs[i-1]
                res = rs[i]
                links = topo.links_to_previous(res)
                if len(links) == 0 or links[0].link_id in ("gap", "?"):
                    con = gemmi.Connection()
                    con.asu = gemmi.Asu.Same
                    con.type = gemmi.ConnectionType.Unknown
                    con.link_id = "gap"
                    con.partner1 = gemmi.AtomAddress(chain.name, res0.seqid, res0.name, "", "\0")
                    con.partner2 = gemmi.AtomAddress(chain.name, res.seqid, res.name, "", "\0")
                    logger.writeln("Refmac workaround (gap link): {}".format(con))
                    st.connections.append(con)

    def fix_microheterogeneity(self, st, topo):
        mh_res = []
        chains = []
        icodes = {} # to avoid overlaps
        modifications = [] # return value

        # Check if microheterogeneity exists
        for chain in st[0]:
            for rg in chain.get_polymer().residue_groups():
                if len(rg) > 1:
                    ress = [r for r in rg]
                    chains.append(chain.name)
                    mh_res.append(ress)
                    ress_str = "/".join([str(r) for r in ress])
                    logger.writeln("Microheterogeneity detected in chain {}: {}".format(chain.name, ress_str))

        if not mh_res: return

        for chain in st[0]:
            for res in chain:
                if res.seqid.icode != " ":
                    icodes.setdefault(chain.name, {}).setdefault(res.seqid.num, []).append(res.seqid.icode)

        def append_links(bond, prr, toappend):
            atoms = bond.atoms
            assert len(atoms) == 2
            found = None
            for i in range(2):
                if any(filter(lambda ra: atoms[i]==ra, prr)): found = i
            if found is not None:
                toappend.append([atoms[i], atoms[1-i]]) # prev atom, current atom
        # append_links()

        mh_res_all = sum(mh_res, [])
        mh_link = {}

        # Check links
        for chain in st[0]:
            for res in chain:
                # If this residue is microheterogeneous
                if res in mh_res_all:
                    for link in topo.links_to_previous(res):
                        mh_link.setdefault(id(res), []).append([link.res1, "prev", link.link_id, []])
                        append_links(topo.first_bond_in_link(link), link.res1, mh_link[id(res)][-1][-1])

                # Check if previous residue(s) is microheterogeneous
                for link in topo.links_to_previous(res):
                    prr = link.res1
                    if prr in mh_res_all:
                        mh_link.setdefault(id(prr), []).append([res, "next", link.link_id, []])
                        append_links(topo.first_bond_in_link(link), prr, mh_link[id(prr)][-1][-1])

        # Change IDs
        for chain_name, rr in zip(chains, mh_res):
            chars = string.ascii_uppercase
            # avoid already used inscodes
            if chain_name in icodes and rr[0].seqid.num in icodes[chain_name]:
                used_codes = set(icodes[chain_name][rr[0].seqid.num])
                chars = list(filter(lambda x: x not in used_codes, chars))
            for ir, r in enumerate(rr[1:]):
                modifications.append([(chain_name, r.seqid.num, r.seqid.icode),
                                      (chain_name, r.seqid.num, chars[ir])])
                r.seqid.icode = chars[ir]

        logger.writeln("DEBUG: mh_link= {}".format(mh_link))
        # Update connections (LINKR)
        for chain_name, rr in zip(chains, mh_res):
            for r in rr:
                for p in mh_link.get(id(r), []):
                    for atoms in p[-1]:
                        con = gemmi.Connection()
                        con.asu = gemmi.Asu.Same
                        con.type = gemmi.ConnectionType.Covale
                        con.link_id = p[2]
                        if p[1] == "prev":
                            p1 = gemmi.AtomAddress(chain_name, p[0].seqid, p[0].name, atoms[1].name, atoms[1].altloc)
                            p2 = gemmi.AtomAddress(chain_name, r.seqid, r.name, atoms[0].name, atoms[0].altloc)
                        else:
                            p1 = gemmi.AtomAddress(chain_name, r.seqid, r.name, atoms[1].name, atoms[1].altloc)
                            p2 = gemmi.AtomAddress(chain_name, p[0].seqid, p[0].name, atoms[0].name, atoms[0].altloc)

                        con.partner1 = p1
                        con.partner2 = p2
                        logger.writeln(" Adding link: {}".format(con))
                        st.connections.append(con)
            for r1, r2 in itertools.combinations(rr, 2):
                for a1 in set([a.altloc for a in r1]):
                    for a2 in set([a.altloc for a in r2]):
                        con = gemmi.Connection()
                        con.asu = gemmi.Asu.Same
                        con.link_id = "gap"
                        # XXX altloc will be ignored when atom does not match.. grrr
                        con.partner1 = gemmi.AtomAddress(chain_name, r1.seqid, r1.name, "", a1)
                        con.partner2 = gemmi.AtomAddress(chain_name, r2.seqid, r2.name, "", a2)
                        st.connections.append(con)

        self.fixes.append(modifications)
    # fix_microheterogeneity()

    def fix_nonpolymer(self, st):
        # Refmac (as of 5.8.0352) has a bug that links non-neighbouring nucleotides
        # It only happens with mmCIF file
        newchains = []
        changes = []
        for chain in st[0]:
            polymer = chain.get_polymer()
            if len(polymer) == len(chain): continue
            if len(polymer) == 0: continue
            del_idxes = []
            newchains.append(gemmi.Chain(self.new_chain_id(chain.name)))
            logger.writeln("Refmac workaround (nonpolymer-fix) {} => {} ({} residues)".format(chain.name, newchains[-1].name,
                                                                                            len(chain) - len(polymer)))
            for i, res in enumerate(chain):
                if res in polymer: continue
                newchains[-1].add_residue(res)
                del_idxes.append(i)
                changes.append([(chain.name, res.seqid.num, res.seqid.icode),
                                (newchains[-1].name, newchains[-1][-1].seqid.num, newchains[-1][-1].seqid.icode)])
            for i in reversed(del_idxes):
                del chain[i]
                
        for c in newchains:
            st[0].add_chain(c)
        if changes:
            st.remove_empty_chains()
            self.fix_metadata(st, dict(changes))
        self.fixes.append(changes)

    def fix_too_large_seqnum(self, st, topo):
        # Refmac cannot handle residue id > 9999
        # What to do:
        # - move to new chains
        # - modify link records (and others?)
        # - add link record if needed
        newchains = []
        changes = []
        
        for chain in st[0]:
            maxseqnum = max([r.seqid.num for r in chain])
            if maxseqnum > self.MAXNUM:
                offset = 0
                #target = [res for res in chain if res.seqid.num > 9999]
                del_idxes = []
                for ires, res in enumerate(chain):
                    if res.seqid.num <= self.MAXNUM: continue
                    if res.seqid.num - offset > self.MAXNUM:
                        newchains.append(gemmi.Chain(self.new_chain_id(chain.name)))
                        offset = res.seqid.num - 1
                        # need to keep link to previous residue if exists
                        for link in topo.links_to_previous(res):
                            logger.writeln("Link: {} {} {} alt= {} {}".format(link.link_id, link.res1, link.res2, 
                                                                            link.alt1, link.alt2))

                            con = gemmi.Connection()
                            con.type = gemmi.ConnectionType.Covale
                            con.link_id = link.link_id
                            #return link
                            bond = topo.first_bond_in_link(link)
                            if bond is not None:
                                con.partner1 = gemmi.AtomAddress(chain.name, link.res1.seqid, link.res1.name, bond.atoms[0].name, bond.atoms[0].altloc)
                                con.partner2 = gemmi.AtomAddress(chain.name, link.res2.seqid, link.res2.name, bond.atoms[1].name, bond.atoms[1].altloc)
                                st.connections.append(con)

                    newchains[-1].add_residue(res)
                    newchains[-1][-1].seqid.num -= offset
                    del_idxes.append(ires)
                    prev = chain[ires-1].seqid if ires > 0 else None
                    changes.append([(chain.name, res.seqid.num, res.seqid.icode),
                                    (newchains[-1].name, newchains[-1][-1].seqid.num, newchains[-1][-1].seqid.icode)])
                    logger.writeln("Refmac workaround (too large seq) {} => {} {}".format(changes[-1][0], changes[-1][1], res.name))

                for i in reversed(del_idxes):
                    del chain[i]

        for c in newchains:
            st[0].add_chain(c)
        if changes:
            st.remove_empty_chains()
            self.fix_metadata(st, dict(changes))
        self.fixes.append(changes)

    def fix_model(self, st, changedict):
        chain_newid = set()
        for chain in st[0]:
            for res in chain:
                changeto = changedict.get((chain.name, res.seqid.num, res.seqid.icode))
                if changeto is not None:
                    logger.writeln("back: {} {} to {}".format(chain.name, res.seqid, changeto))
                    #chain.name = changeto[0] # this is ok when modify back
                    chain_newid.add((chain, changeto[0]))
                    res.seqid.num = changeto[1]
                    res.seqid.icode = changeto[2]

        for chain, newid in chain_newid:
            chain.name = newid
        st.merge_chain_parts()
        self.fix_metadata(st, changedict)
        
    def modify_back(self, st):
        for fix in reversed(self.fixes):
            reschanges = dict([x[::-1] for x in fix])
            self.fix_model(st, reschanges)
        
        
class Refmac:
    def __init__(self, **kwargs):
        self.prefix = "refmac"
        self.hklin = self.xyzin = ""
        self.source = "electron"
        self.lab_f = None
        self.lab_sigf = None
        self.lab_phi = None
        self.libin = None
        self.tlsin = None
        self.hydrogen = "all"
        self.hout = False
        self.ncycle = 10
        self.tlscycle = 0
        self.resolution = None
        self.weight_matrix = None
        self.weight_auto_scale = None
        self.bfactor = None
        self.jellybody = None
        self.jellybody_sigma, self.jellybody_dmax = 0.01, 4.2
        self.ncsr = None
        self.shake = None
        self.keyword_files = []
        self.keywords = []
        self.external_restraints_json = None
        self.exe = "refmac5"
        self.monlib_path = None
        self.keep_chain_ids = False
        self.show_log = False # summary only if false
        self.global_mode = kwargs.get("global_mode")
        
        for k in kwargs:
            if k == "args":
                self.init_from_args(kwargs["args"])
            else:
                setattr(self, k, kwargs[k])

        ensure_ccp4scr()
    # __init__()

    def init_from_args(self, args):
        self.hklin = args.mtz
        self.xyzin = args.model
        self.libin = args.ligand
        self.tlsin = args.tlsin
        self.ncycle = args.ncycle
        self.tlscycle = args.tlscycle
        self.lab_f = args.lab_f
        self.lab_phi = args.lab_phi
        self.lab_sigf = args.lab_sigf
        self.hydrogen = args.hydrogen
        self.hout = args.hout
        self.ncsr = args.ncsr
        self.bfactor = args.bfactor
        self.jellybody = args.jellybody
        self.jellybody_sigma, self.jellybody_dmax = args.jellybody_params
        self.resolution = args.resolution
        self.weight_auto_scale = args.weight_auto_scale
        self.keyword_files = args.keyword_file
        self.keywords = args.keywords
        self.external_restraints_json = args.external_restraints_json
        self.exe = args.exe
        self.show_log = args.show_refmac_log
        self.monlib_path = args.monlib
    # init_from_args()

    def copy(self, **kwargs):
        ret = copy.deepcopy(self)
        for k in kwargs:
            setattr(ret, k, kwargs[k])
            
        return ret
    # copy()

    def set_libin(self, ligands):
        if not ligands: return
        if len(ligands) > 1:
            mcif = "merged_ligands.cif" # XXX directory!
            logger.writeln("Merging ligand cif files: {}".format(ligands))
            fileio.merge_ligand_cif(ligands, mcif)
            self.libin = mcif
        else:
            self.libin = ligands[0]
    # set_libin()
            
    
    def make_keywords(self):
        ret = ""
        labin = []
        if self.lab_f: labin.append("FP={}".format(self.lab_f))
        if self.lab_sigf: labin.append("SIGFP={}".format(self.lab_sigf))
        if self.lab_phi: labin.append("PHIB={}".format(self.lab_phi))
        if labin:
            ret += "labin {}\n".format(" ".join(labin))


        ret += "make hydr {}\n".format(self.hydrogen)
        ret += "make hout {}\n".format("yes" if self.hout else "no")
        
        if self.global_mode == "spa":
            ret += "solvent no\n"
            ret += "scale lssc isot\n"
            ret += "source em mb\n"
        elif self.source == "electron":
            ret += "source ec mb\n"
        elif self.source == "neutron":
            ret += "source n\n"

        ret += "ncycle {}\n".format(self.ncycle)
        if self.resolution is not None:
            ret += "reso {}\n".format(self.resolution)
        if self.weight_matrix is not None:
            ret += "weight matrix {}\n".format(self.weight_matrix)
        elif self.weight_auto_scale is not None:
            ret += "weight auto {:.2e}\n".format(self.weight_auto_scale)
        else:
            ret += "weight auto\n"

        if self.bfactor is not None:
            ret += "bfactor set {}\n".format(self.bfactor)
        if self.jellybody:
            ret += "ridge dist sigma {:.3e}\n".format(self.jellybody_sigma)
            ret += "ridge dist dmax {:.2e}\n".format(self.jellybody_dmax)
        if self.ncsr:
            ret += "ncsr {}\n".format(self.ncsr)
        if self.shake:
            ret += "rand {}\n".format(self.shake)
        if self.tlscycle > 0:
            ret += "refi tlsc {}\n".format(self.tlscycle)
            ret += "tlsout addu\n"
        if self.keep_chain_ids:
            ret += "pdbo keep auth\n"

        if self.external_restraints_json:
            ret += external_restraints_json_to_keywords(self.external_restraints_json)
            
        if self.keyword_files:
            for f in self.keyword_files:
                ret += "@{}\n".format(f)

        if self.keywords:
            ret += "\n".join(self.keywords).strip() + "\n"

        return ret
    # make_keywords()

    def xyzout(self): return self.prefix + ".pdb"
    def hklout(self): return self.prefix + ".mtz"
    def tlsout(self): return self.prefix + ".tls"

    def make_cmd(self):
        cmd = [self.exe]
        cmd.extend(["hklin", self.hklin])
        cmd.extend(["hklout", self.hklout()])
        cmd.extend(["xyzin", self.xyzin])
        cmd.extend(["xyzout", self.xyzout()])
        if self.libin:
            cmd.extend(["libin", self.libin])
        if self.tlsin:
            cmd.extend(["tlsin", self.tlsin])
        if self.tlscycle > 0:
            cmd.extend(["tlsout", self.tlsout()])
        if self.source == "neutron":
            cmd.extend(["atomsf", os.path.join(os.environ["CLIBD"], "atomsf_neutron.lib")])
            
        return cmd
    # make_cmd()

    def run_refmac(self, write_summary_json=True):
        cmd = self.make_cmd()
        stdin = self.make_keywords()
        with open(self.prefix+".inp", "w") as ofs: ofs.write(stdin)

        logger.writeln("Running REFMAC5..")
        logger.writeln("{} <<__eof__ > {}".format(" ".join(pipes.quote(x) for x in cmd), self.prefix+".log"))
        logger.write(stdin)
        logger.writeln("__eof__")

        env = os.environ
        if self.monlib_path: env["CLIBD_MON"] = os.path.join(self.monlib_path, "") # should end with /
        
        p = subprocess.Popen(cmd, shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                             universal_newlines=True, env=env)
        p.stdin.write(stdin)
        p.stdin.close()

        log = open(self.prefix+".log", "w")
        cycle = 0
        re_cycle_table = re.compile("Cycle *([0-9]+). Rfactor analysis")
        re_actual_weight = re.compile("Actual weight *([^ ]+) *is applied to the X-ray term")
        rmsbond = ""
        rmsangle = ""
        log_delay = []
        summary_write = (lambda x: log_delay.append(x)) if self.show_log else logger.writeln
        outlier_flag = False
        last_table_flag = False
        last_table_keys = []
        occ_flag = False
        occ_cycles = 0
        ret = {"version":None,
               "cycles": [{"cycle":i} for i in range(self.ncycle+self.tlscycle+1)],
               } # metadata
        
        for l in iter(p.stdout.readline, ""):
            log.write(l)

            if self.show_log:
                print(l, end="")
            
            r_ver = re_version.search(l)
            if r_ver:
                ret["version"] = r_ver.group(1)
                summary_write("Starting Refmac {} (PID: {})".format(r_ver.group(1), p.pid))

            # print error/warning
            r_err = re_error.search(l)
            if r_err:
                if self.global_mode == "spa":
                    if "Figure of merit of phases has not been assigned" in l:
                        continue
                    elif "They will be assumed to be equal to 1.0" in l:
                        continue
                summary_write(l.rstrip())

            # print outliers
            r_outl = re_outlier_start.search(l)
            if r_outl:
                outlier_flag = True
                summary_write(l.rstrip())
            elif outlier_flag:
                if l.strip() == "" or "monitored" in l or "dev=" in l or "sigma=" in l.lower() or "sigma.=" in l:
                    summary_write(l.rstrip())
                else:
                    outlier_flag = False

            if "TLS refinement cycle" in l:
                cycle = int(l.split()[-1])
            elif "----Group occupancy refinement----" in l:
                occ_flag = True
                occ_cycles += 1
                cycle += 1
            elif "CGMAT cycle number =" in l:
                cycle = int(l[l.index("=")+1:]) + self.tlscycle + occ_cycles
                occ_flag = False
            
            r_cycle = re_cycle_table.search(l)
            if r_cycle: cycle = int(r_cycle.group(1))                

            for i in range(len(ret["cycles"]), cycle):
                ret["cycles"].append({"cycle":i})

            if "Overall R factor                     =" in l and cycle > 0:
                rfac = l[l.index("=")+1:].strip()
                if self.global_mode != "spa":
                    ret["cycles"][cycle-1]["r_factor"] = rfac
                    summary_write(" cycle= {:3d} R= {}".format(cycle-1, rfac))
            elif "Average Fourier shell correlation    =" in l and cycle > 0:
                fsc = l[l.index("=")+1:].strip()
                if occ_flag:
                    note = "(occupancy)"
                elif cycle == 1:
                    note = "(initial)"
                elif cycle <= self.tlscycle+1:
                    note = "(TLS)"
                elif cycle > self.ncycle + occ_cycles + self.tlscycle:
                    note = "(final)"
                else:
                    note = ""

                if self.global_mode == "spa":
                    ret["cycles"][cycle-1]["fsc_average"] = fsc
                    summary_write(" cycle= {:3d} FSCaverage= {} {}".format(cycle-1, fsc, note))
            elif "Rms BondLength" in l:
                rmsbond = l
            elif "Rms BondAngle" in l:
                rmsangle = l

            r_actual_weight = re_actual_weight.search(l)
            if r_actual_weight:
                ret["cycles"][cycle-1]["actual_weight"] = r_actual_weight.group(1)

            # Final table
            if "    Ncyc    Rfact    Rfree     FOM" in l:
                last_table_flag = True
                last_table_keys = l.split()
                if last_table_keys[-1] == "$$": del last_table_keys[-1]
            elif last_table_flag:
                if "$$ Final results $$" in l:
                    last_table_flag = False
                    continue
                sp = l.split()
                if len(sp) == len(last_table_keys) and sp[0] != "$$":
                    cyc = int(sp[last_table_keys.index("Ncyc")])
                    key_name = dict(rmsBOND="rms_bond", zBOND="rmsz_bond",
                                    rmsANGL="rms_angle", zANGL="rmsz_angle",
                                    rmsCHIRAL="rms_chiral")
                    for k in key_name:
                        if k in last_table_keys:
                            ret["cycles"][cyc][key_name[k]] = sp[last_table_keys.index(k)]
                        else:
                            logger.error("table does not have key {}?".format(k))
                            
        retcode = p.wait()
        log.close()
        if log_delay:
            logger.writeln("== Summary of Refmac ==")
            logger.writeln("\n".join(log_delay))
                         
        if rmsbond:
            logger.writeln("                      Initial    Final")
            logger.writeln(rmsbond.rstrip())
            logger.writeln(rmsangle.rstrip())
            
        logger.writeln("REFMAC5 finished with exit code= {}".format(retcode))

        if write_summary_json:
            json.dump(ret,
                      open("{}_summary.json".format(self.prefix), "w"),
                      indent=True)

        # TODO check timestamp
        if not os.path.isfile(self.xyzout()) or not os.path.isfile(self.hklout()):
            raise RuntimeError("REFMAC5 did not produce output files. Check {}".format(self.prefix+".log"))

        return ret
    # run_refmac()
# class Refmac



# Care starting grid!!
"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import os
import gemmi
import numpy
import argparse
import json
from servalcat.utils import logger
from servalcat import utils

def add_arguments(parser):
    parser.description = 'Trim maps and shift models into a small new box.'
    parser.epilog = 'If --mask is provided, a boundary is decided using the mask and --padding. Otherwise the model is used.'
    parser.add_argument('--maps', nargs="+", action="append",
                        help='Input map file(s)')
    parser.add_argument('--mask',
                        help='Mask file')    
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    parser.add_argument('--model', nargs="+", action="append",
                        help='Input atomic model file(s)')
    parser.add_argument("--no_expand_ncs", action='store_true',
                        help="Do not expand strict NCS in MTRIX or _struct_ncs_oper")
    parser.add_argument('--padding', type=float, default=10.0,
                        help='padding in angstrom unit (default: %(default).1f)')
    parser.add_argument('--mask_cutoff', type=float, default=0.5,
                        help='Mask value cutoff to define boundary (default: %(default).1f)')
    parser.add_argument('--noncubic',
                        action='store_true')
    parser.add_argument('--noncentered',
                        action='store_true',
                        help='If specified non-centered trimming is performed. Not recommended if having some symmetry')
    parser.add_argument('--no_shift',
                        action='store_true',
                        help='If specified resultant maps will have shifted origin and overlap with the input maps.')
    parser.add_argument('--no_shift_keep_cell',
                        action='store_true',
                        help='Keep original unit cell when --no_shift is given')
    parser.add_argument('--force_cell', type=float, nargs=6,
                        metavar=("a", "b", "c", "alpha", "beta", "gamma"),
                        help='Use specified unit cell parameter')
    parser.add_argument('--disable_cell_check', action='store_true',
                        help="Turn off unit cell consistency test")
    parser.add_argument("--shifts", help="Specify shifts.json to use precalculated parameters")
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def check_maps(map_files, pixel_size=None, disable_cell_check=False):
    logger.writeln("Input map files:")
    params = []
    for f in map_files:
        g, gs = utils.fileio.read_ccp4_map(f, pixel_size=pixel_size)
        params.append((g.unit_cell.parameters, g.shape, g.spacing, gs))

    shapes = set([x[1] for x in params])
    if len(shapes) > 1:
        raise RuntimeError("Error: different grid size included")

    starts = set([tuple(x[3]) for x in params])
    if len(starts) > 1:
        raise RuntimeError("Error: different origins included")
    
    cells = set([x[0] for x in params])
    if len(cells) > 1:
        if disable_cell_check:
            logger.writeln("WARNING: Cells are different. Using the first one.")
        else:
            raise RuntimeError("Error: different cell parameters included")

    return params[0]
# check_maps()

def write_shifts_json(filename, cell, shape, new_cell, new_shape, starts, shifts):
    json.dump(dict(cell=cell,
                   grid=shape,
                   new_cell=new_cell,
                   new_grid=new_shape,
                   starts=starts,
                   shifts=shifts),
              open(filename, "w"), indent=2)
# write_shifts_json()

def determine_shape_and_shift(mask, grid_start, padding, sts=None, mask_cutoff=1e-5, noncentered=False, noncubic=False,
                              json_out="trim_shifts.json"):
    # XXX need to think more carefully.
    #     probably we need a special class for non-crystallographic maps (reset unit cell with box size)
    #     negative grid_start with mask (not model) may be problematic
    grid_shape = numpy.array(mask.shape)
    spacing = numpy.array(mask.spacing)
    grid_start = numpy.array(grid_start)
    grid_end = grid_start + grid_shape # for indexing grid_end-1 is the end
    cell = mask.unit_cell
    logger.writeln("Original grid start: {:4d} {:4d} {:4d}".format(*grid_start))
    logger.writeln("         grid   end: {:4d} {:4d} {:4d}".format(*(grid_end-1)))
    if sts:
        # here model is used to define region and mask content is ignored.
        allpos = numpy.array([cra.atom.pos.tolist() for st in sts for cra in st[0].all()])
        minp = mask.get_nearest_point(gemmi.Position(*numpy.min(allpos, axis=0)))
        maxp = mask.get_nearest_point(gemmi.Position(*numpy.max(allpos, axis=0)))
        tmp = [(minp.u, maxp.u), (minp.v, maxp.v), (minp.w, maxp.w)] - grid_start[:,None]
    else:
        tmp = numpy.where(mask.array>mask_cutoff) - grid_start[:,None]
    
    tmp += (grid_shape[:,None]*numpy.floor(1-tmp/grid_shape[:,None])).astype(int) + grid_start[:,None]
    limits = [(min(x), max(x)) for x in tmp]

    p = numpy.ceil(padding / spacing).astype(int)
    logger.writeln("Limits: {} {} {}".format(*limits))
    logger.writeln("Padding: {} {} {}".format(*p))
    slices = [0, 0, 0]
    if noncentered:
        logger.writeln("Non-centered trimming will be performed.")
        for i in range(3):
            start = max(grid_start[i], limits[i][0]-p[i])
            stop = min(limits[i][1]+p[i]+1, grid_end[i])
            if (stop-start)%2 == 1:
                if start > 0: start -= 1
                elif stop < grid_end[i]: stop += 1
            slices[i] = slice(start, stop)
    else:
        logger.writeln("Centered trimming will be performed.")
        for i in range(3):
            ctr = (grid_shape[i]-1)/2 + grid_start[i]
            rad = max(ctr-(limits[i][0]-p[i]), (limits[i][1]+p[i])-ctr)
            logger.writeln("Rad{}= {}".format(i, rad))
            if rad < grid_shape[i]/2:
                slices[i] = slice(int(ctr-rad), int(ctr+rad)+1, None)
            else:
                slices[i] = slice(grid_start[i], grid_end[i], None)
            
    logger.writeln("Slices: {}".format(slices))
    if not noncubic:
        # only works when input grid is cubic; otherwise need to expand
        if len(set(mask.shape)) > 1:
            raise RuntimeError("Input grid is not cubic. Try --noncubic")
        min_s = min([slices[i].start for i in range(3)])
        max_s = max([slices[i].stop for i in range(3)])
        slices = [slice(min_s, max_s, None) for i in range(3)]
        logger.writeln("Cubic Slices: {}".format(slices))

    starts = [slices[i].start for i in range(3)]
        
    # Shifts for model
    shifts = -mask.get_position(slices[0].start, slices[1].start, slices[2].start)
    logger.writeln("Shift for model: {} {} {}".format(*shifts.tolist()))

    new_shape = [slices[0].stop-slices[0].start,
                 slices[1].stop-slices[1].start,
                 slices[2].stop-slices[2].start]
    tmp = mask.get_position(*new_shape)
    new_cell = gemmi.UnitCell(tmp[0], tmp[1], tmp[2], cell.alpha, cell.beta, cell.gamma)
    
    logger.writeln("New Cell: {:.4f} {:.4f} {:.4f} {:.3f} {:.3f} {:.3f}".format(*new_cell.parameters))
    logger.writeln("New grid: {} {} {}".format(*new_shape))

    if json_out:
        write_shifts_json(json_out,
                          cell=mask.unit_cell.parameters, shape=mask.shape,
                          new_cell=new_cell.parameters, new_shape=list(map(int, new_shape)),
                          starts=list(map(int, starts)),
                          shifts=shifts.tolist())

    return new_cell, new_shape, starts, shifts
# determine_shape_and_shift()    
        
def main(args):
    if not args.maps and not args.shifts:
        raise SystemExit("Give --maps or --shifts")

    if not args.mask and args.model and not args.shifts and args.padding <= 0:
        raise SystemExit("--padding must be > 0 if you want to create a mask from the model.")

    if args.no_shift_keep_cell and not args.no_shift:
        logger.writeln("WARNING: --no_shift_keep_cell has no effect when --no_shift not given")
    
    if args.maps:
        args.maps = sum(args.maps, [])
    else:
        args.maps = []
    
    if args.shifts:
        if args.noncubic or args.noncentered or args.mask:
            raise SystemExit("You cannot specify --noncubic/--noncentered/--mask if --shifts given")
        if not args.maps and not args.model:
            raise SystemExit("Give --maps or --model")
        info = json.load(open(args.shifts))
        cell = info["cell"]
    elif args.maps:
        cell, grid_shape, spacing, grid_start = check_maps(args.maps, args.pixel_size, args.disable_cell_check)
      
    if args.force_cell:
        cell = args.force_cell

    cell = gemmi.UnitCell(*cell)
    sts, cif_refs = [], []
    sts_for_mask = []
    if args.model:
        args.model = sum(args.model, [])
        for m in args.model:
            st, cif_ref = utils.fileio.read_structure_from_pdb_and_mmcif(m)
            st.spacegroup_hm = "P1"
            st.cell = cell
            # XXX here we do not want moving into box (with "crystallographic" translation) at least for --no_shift
            sts.append(st)
            cif_refs.append(cif_ref)
        
    if args.mask:
        logger.writeln("Using mask to decide border: {}".format(args.mask))
        mask = utils.fileio.read_ccp4_map(args.mask)[0]
        assert mask.shape == grid_shape
        mask.set_unit_cell(cell)
        mask.spacegroup = gemmi.SpaceGroup(1)
        if args.mask not in args.maps: # need to check with normalized paths?
            args.maps.append(args.mask)
    elif args.model and not args.shifts:
        logger.writeln("Using model to decide border: {}".format(args.model))
        mask = gemmi.FloatGrid(*grid_shape) # this is just dummy. values will not be seen.
        mask.set_unit_cell(cell)
        mask.spacegroup = gemmi.SpaceGroup(1)
        for st in sts:
            if len(st.ncs) > 0 and not args.no_expand_ncs:
                st = st.clone()
                utils.model.expand_ncs(st)
            sts_for_mask.append(st)
    elif not args.shifts:
        raise SystemExit("Give mask or model")

    if not args.shifts:
        new_cell, new_shape, starts, shifts = determine_shape_and_shift(mask=mask, grid_start=grid_start,
                                                                        sts=sts_for_mask,
                                                                        padding=args.padding,
                                                                        mask_cutoff=args.mask_cutoff,
                                                                        noncentered=args.noncentered,
                                                                        noncubic=args.noncubic)
    else:
        new_cell = gemmi.UnitCell(*info["new_cell"])
        new_shape, starts = info["new_grid"], info["starts"]
        shifts = gemmi.Position(*info["shifts"])
        
    if args.model and not args.no_shift:
        for i, st in enumerate(sts):
            st.cell = new_cell
            # apply shifts to model
            for mol in st:
                for cra in mol.all():
                    cra.atom.pos += shifts

            # apply shifts to translations
            spext = utils.fileio.splitext(os.path.basename(args.model[i]))
            if len(st.ncs) > 0 and not args.no_expand_ncs:
                new_ops = utils.symmetry.apply_shift_for_ncsops(st.ncs, shifts)
                st.ncs.clear()
                st.ncs.extend(new_ops)
                logger.writeln(" Writing symmetry expanded model for shifted model")
                utils.symmetry.write_symmetry_expanded_model(st, spext[0]+"_trimmed_local_expanded",
                                                             pdb=True, cif=True)

            # save files
            utils.fileio.write_model(st, file_name=spext[0]+"_trimmed"+spext[1], cif_ref=cif_refs[i])
    
    for f in args.maps:
        logger.writeln("Slicing {}".format(f))
        outf = os.path.basename(utils.fileio.splitext(f)[0])+"_trimmed.mrc"
        g = utils.fileio.read_ccp4_map(f, pixel_size=args.pixel_size)[0]
        if args.no_shift:
            utils.maps.write_ccp4_map(outf, g, cell=cell, sg=g.spacegroup,
                                      grid_start=starts, grid_shape=new_shape,
                                      update_cell=not args.no_shift_keep_cell)
        else:
            newg = g.get_subarray(starts, new_shape)
            utils.maps.write_ccp4_map(outf, newg, cell=new_cell, sg=g.spacegroup)

# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import pandas
import os
import argparse
from servalcat.utils import logger
from servalcat import utils

def add_arguments(parser):
    parser.description = 'Calculate real space local correlation map from half maps and model'
    parser.add_argument("--halfmaps", required=True, nargs=2,
                       help="Input half map files")
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--kernel", type=int, 
                       help="Kernel radius in pixel")
    group.add_argument("--kernel_ang", type=float,
                       help="Kernel radius in Angstrom (hard sphere)")
    parser.add_argument('--mask',
                        help="mask file")
    parser.add_argument('--model', 
                        help='Input atomic model file')
    parser.add_argument('--resolution', type=float,
                        help='default: nyquist resolution')
    parser.add_argument("--trim", action='store_true', help="Write trimmed map")
    parser.add_argument('-o', '--output_prefix', default="ccmap",
                        help="default: %(default)s")
# add_arguments()
                        
def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def setup_coeffs_for_halfmap_cc(maps, d_min, mask=None, st=None):
    hkldata = utils.maps.mask_and_fft_maps(maps, d_min, mask)
    hkldata.setup_relion_binning()
    utils.maps.calc_noise_var_from_halfmaps(hkldata)

    nref = len(hkldata.df.index)
    F1w = numpy.zeros(nref, dtype=complex)
    F2w = numpy.zeros(nref, dtype=complex)
    F1 = hkldata.df.F_map1.to_numpy()
    F2 = hkldata.df.F_map2.to_numpy()

    logger.writeln("Calculating weights for half map correlation.")
    logger.writeln(" weight = sqrt(FSChalf / (2*var_noise + var_signal))")
    hkldata.binned_df["w2_half_varsignal"] = 0.
    for i_bin, idxes in hkldata.binned():
        fscfull = hkldata.binned_df.FSCfull[i_bin]
        if fscfull < 0:
            break # stop here so that higher resolution are all zero
        fsc = fscfull / (2 - fscfull)
        var_fo = 2 * hkldata.binned_df.var_noise[i_bin] + hkldata.binned_df.var_signal[i_bin]
        w = numpy.sqrt(fsc / var_fo)
        hkldata.binned_df.loc[i_bin, "w2_half_varsignal"] = fsc / var_fo * hkldata.binned_df.var_signal[i_bin]
        F1w[idxes] = F1[idxes] * w
        F2w[idxes] = F2[idxes] * w

    hkldata.df["F_map1w"] = F1w
    hkldata.df["F_map2w"] = F2w

    return hkldata
# setup_coeffs_for_halfmap_cc()

def add_coeffs_for_model_cc(hkldata, st):
    hkldata.df["FC"] = utils.model.calc_fc_fft(st, d_min=hkldata.d_min_max()[0]-1e-6,
                                               source="electron", miller_array=hkldata.miller_array())
    nref = len(hkldata.df.index)
    FCw = numpy.zeros(nref, dtype=complex)
    FPw = numpy.zeros(nref, dtype=complex)
    FP = hkldata.df.FP.to_numpy()
    FC = hkldata.df.FC.to_numpy()

    logger.writeln("Calculating weights for map-model correlation.")
    logger.writeln(" weight for Fo = sqrt(FSCfull / var(Fo))")
    logger.writeln(" weight for Fc = sqrt(FSCfull / var(Fc))")
    hkldata.binned_df["w_mapmodel_c"] = 0.
    hkldata.binned_df["w_mapmodel_o"] = 0.
    hkldata.binned_df["var_fc"] = 0.
    for i_bin, idxes in hkldata.binned():
        fscfull = hkldata.binned_df.FSCfull[i_bin]
        if fscfull < 0: break
        var_fc = numpy.var(FC[idxes])
        wc = numpy.sqrt(fscfull / var_fc)
        wo = numpy.sqrt(fscfull / numpy.var(FP[idxes]))
        FCw[idxes] = FC[idxes] * wc
        FPw[idxes] = FP[idxes] * wo
        hkldata.binned_df.loc[i_bin, "w_mapmodel_c"] = wc
        hkldata.binned_df.loc[i_bin, "w_mapmodel_o"] = wo
        hkldata.binned_df.loc[i_bin, "var_fc"] = var_fc
    
    hkldata.df["FPw"] = FPw
    hkldata.df["FCw"] = FCw
# add_coeffs_for_model_cc()

def model_stats(st, modelcc_map, halfcc_map, loggraph_out=None):
    tmp = dict(chain=[], seqid=[], resn=[], CC_mapmodel=[], CC_halfmap=[])
    for chain in st[0]:
        for res in chain:
            mm = numpy.mean([modelcc_map.interpolate_value(atom.pos) for atom in res])
            hc = numpy.mean([halfcc_map.interpolate_value(atom.pos) for atom in res])
            tmp["chain"].append(chain.name)
            tmp["seqid"].append(str(res.seqid))
            tmp["resn"].append(res.name)
            tmp["CC_mapmodel"].append(mm)
            tmp["CC_halfmap"].append(hc)

    df = pandas.DataFrame(tmp)
    df["sqrt_CC_full"] = numpy.sqrt(2 * df.CC_halfmap / (1 + df.CC_halfmap))
    if loggraph_out is not None:
        with open(loggraph_out, "w") as ofs:
            for c, g in df.groupby("chain", sort=False):
                ofs.write("$TABLE: Chain {} :".format(c))
                ofs.write("""
$GRAPHS
: average correlations :A:2,4,5,6:
$$
chain seqid resn CC(map,model) CC_half sqrt(CC_full)
$$
$$
""")
                ofs.write(g.to_string(header=False, index=False))
                ofs.write("\n\n")
    return df
# model_stats()

def main(args):
    maps = utils.fileio.read_halfmaps(args.halfmaps, pixel_size=args.pixel_size)
    grid_shape = maps[0][0].shape
    if args.mask:
        mask = utils.fileio.read_ccp4_map(args.mask)[0]
    else:
        mask = None

    if args.resolution is None:
        d_min = utils.maps.nyquist_resolution(maps[0][0])
    else:
        d_min = args.resolution

    hkldata = setup_coeffs_for_halfmap_cc(maps, d_min, mask)
    if args.kernel is None:
        prefix = "{}_r{}A".format(args.output_prefix, args.kernel_ang)
        knl = hkldata.hard_sphere_kernel(r_ang=args.kernel_ang, grid_size=grid_shape)
    else:
        prefix = "{}_r{}px".format(args.output_prefix, args.kernel)
        knl = utils.maps.raised_cosine_kernel(args.kernel)
        
    halfcc_map = utils.maps.local_cc(hkldata.fft_map("F_map1w", grid_size=grid_shape),
                                     hkldata.fft_map("F_map2w", grid_size=grid_shape),
                                     knl, method="simple" if args.kernel is None else "scipy")

    halfcc_map_in_mask = halfcc_map.array[mask.array>0.5] if mask is not None else halfcc_map
    logger.writeln("Half map CC: min/max= {:.4f} {:.4f}".format(numpy.min(halfcc_map_in_mask), numpy.max(halfcc_map_in_mask)))
    utils.maps.write_ccp4_map(prefix+"_half.mrc", halfcc_map, hkldata.cell, hkldata.sg,
                              mask_for_extent=mask if args.trim else None)

    if args.model:
        st = utils.fileio.read_structure(args.model)
        utils.model.expand_ncs(st)
        st.cell = hkldata.cell
        st.spacegroup_hm = hkldata.sg.hm
        add_coeffs_for_model_cc(hkldata, st)
        modelcc_map = utils.maps.local_cc(hkldata.fft_map("FPw", grid_size=grid_shape),
                                          hkldata.fft_map("FCw", grid_size=grid_shape),
                                          knl, method="simple" if args.kernel is None else "scipy")
        modelcc_map_in_mask = modelcc_map.array[mask.array>0.5] if mask is not None else modelcc_map
        logger.writeln("Model-map CC: min/max= {:.4f} {:.4f}".format(numpy.min(modelcc_map_in_mask), numpy.max(modelcc_map_in_mask)))
        utils.maps.write_ccp4_map(prefix+"_model.mrc", modelcc_map, hkldata.cell, hkldata.sg,
                                  mask_for_extent=mask if args.trim else None)
        model_stats(st, modelcc_map, halfcc_map, loggraph_out=prefix+"_byresidue.log")
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import json
import os
import argparse
from servalcat.utils import logger
from servalcat import utils

def add_arguments(parser):
    parser.description = 'Run REFMAC5 for small molecule crystallography'
    parser.add_argument('--exe', default="refmac5", help='refmac5 binary')
    parser.add_argument('--cif', help="cif file containing model and data")
    parser.add_argument('--sg', help="Space group")
    parser.add_argument('--model', 
                        help='Input atomic model file')
    parser.add_argument('--hklin',
                        help='Input reflection file')
    parser.add_argument('--hklin_labs', nargs='+',
                        help='column names to be used')
    parser.add_argument('--blur', type=float,
                        help='Apply B-factor blurring to --hklin')
    parser.add_argument('--resolution',
                        type=float,
                        help='')
    parser.add_argument('--ligand', nargs="*", action="append",
                        help="restraint dictionary cif file(s)")
    parser.add_argument('--ncycle', type=int, default=10)
    #parser.add_argument('--jellybody', action='store_true')
    #parser.add_argument('--jellybody_params', nargs=2, type=float,
    #                    metavar=("sigma", "dmax"), default=[0.01, 4.2])
    parser.add_argument('--hydrogen', default="all", choices=["all", "yes", "no"],
                        help="all: add riding hydrogen atoms, yes: use hydrogen atoms if present, no: remove hydrogen atoms in input")
    parser.add_argument('--no_hout', action='store_true', help="do not write hydrogen atoms in the output model")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--weight_auto_scale', type=float,
                       help="'weight auto' scale value. automatically determined from resolution and mask/box volume ratio if unspecified")
    group.add_argument('--weight_matrix', type=float,
                       help="weight matrix value")
    parser.add_argument('--invert', action='store_true', help="invert handednes")
    parser.add_argument('--bref', choices=["aniso","iso","iso_then_aniso"], default="aniso")
    parser.add_argument('--unrestrained', action='store_true')
    parser.add_argument('--bulk_solvent', action='store_true')
    parser.add_argument('-s', '--source', choices=["electron", "xray", "neutron"], default="electron") #FIXME
    parser.add_argument('--keywords', nargs='+', action="append",
                        help="refmac keyword(s)")
    parser.add_argument('--keyword_file', nargs='+', action="append",
                        help="refmac keyword file(s)")
    parser.add_argument('--external_restraints_json')
    parser.add_argument('--show_refmac_log', action='store_true')
    parser.add_argument('--output_prefix', default="refined",
                        help='output file name prefix')
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
# add_arguments()
                        
def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def write_mtz(mtz_out, asudata, hklf, blur=None):
    data = numpy.hstack((asudata.miller_array,
                         asudata.value_array["value"][:,numpy.newaxis],
                         asudata.value_array["sigma"][:,numpy.newaxis]))
    mtz = gemmi.Mtz()
    mtz.spacegroup = asudata.spacegroup
    mtz.cell = asudata.unit_cell
    mtz.add_dataset('HKL_base')
    for label in ['H', 'K', 'L']: mtz.add_column(label, 'H')

    if hklf == 3:
        mtz.add_column("F", "F")
        mtz.add_column("SIGF", "Q")
    else:
        mtz.add_column("I", "J")
        mtz.add_column("SIGI", "Q")

    mtz.set_data(data)
    if blur is not None: utils.hkl.blur_mtz(mtz, blur)
    mtz.write_to_file(mtz_out)
# write_mtz()

def make_invert_tr(sg, cell):
    ops = sg.operations()
    coh = sg.change_of_hand_op()
    ops.change_basis_forward(sg.change_of_hand_op())
    new_sg = gemmi.find_spacegroup_by_ops(ops)
    tr = cell.op_as_transform(coh)
    return new_sg, tr
# make_invert_tr()
    
def main(args):
    if not args.cif and not (args.model and args.hklin):
        raise SystemExit("Give [--model and --hklin] or --cif")

    if args.sg:
        try:
            sg_user = gemmi.SpaceGroup(args.sg)
            logger.writeln("User-specified space group: {}".format(sg_user.xhm()))
        except ValueError:
            raise SystemExit("Error: Unknown space group '{}'".format(args.sg))
    else:
        sg_user = None

    mtz_in = "input.mtz" # always write this file as an input for Refmac
    if args.cif:
        asudata, ss, info = utils.fileio.read_smcif_shelx(args.cif)
        st = utils.model.cx_to_mx(ss)
        if sg_user:
            if not asudata.unit_cell.is_compatible_with_spacegroup(sg_user):
                raise SystemExit("Error: Specified space group {} is incompatible with the unit cell parameters {}".format(sg_user.xhm(),
                                                                                                                           asudata.unit_cell.parameters))
            
            asudata.spacegroup = sg_user
        write_mtz(mtz_in, asudata, info.get("hklf"), args.blur)
    else:
        st = utils.fileio.read_small_structure(args.model)
        sg_st = st.find_spacegroup()
        logger.writeln(" Cell from model: {}".format(st.cell))
        logger.writeln(" Space group from model: {}".format(st.spacegroup_hm))

        if args.hklin.endswith(".mtz"): # TODO may be unmerged mtz
            logger.writeln("Reading MTZ file: {}".format(args.hklin))
            mtz = gemmi.read_mtz_file(args.hklin)
            logger.writeln(" Cell from mtz: {}".format(mtz.cell))
            if not mtz.cell.approx(st.cell, 1e-3):
                logger.writeln(" Warning: unit cell mismatch!")
            logger.writeln(" Space group from mtz: {}".format(mtz.spacegroup.hm))
            if sg_user:
                if not mtz.cell.is_compatible_with_spacegroup(sg_user):
                    raise SystemExit("Error: Specified space group {} is incompatible with the unit cell parameters {}".format(sg_user.xhm(),
                                                                                                                               mtz.cell.parameters))
                mtz.spacegroup = sg_user
                logger.writeln(" Writing {} as space group {}".format(mtz_in, sg_user.xhm()))
            elif mtz.spacegroup != sg_st:
                logger.writeln(" Warning: space group mismatch between model and mtz")
                if mtz.cell.is_compatible_with_spacegroup(sg_st):
                    logger.writeln("  Taking space group from model.")
                    mtz.spacegroup = sg_st
                    logger.writeln("  Writing {} as space group {}".format(mtz_in, sg_st.xhm()))
                else:
                    logger.writeln("  Model space group is incompatible with mtz unit cell. Ignoring model space group.")

            if args.hklin_labs:
                try: mtz = utils.hkl.mtz_selected(mtz, args.hklin_labs)
                except RuntimeError as e:
                    raise SystemExit("Error: {}".format(e))
            if args.blur is not None: utils.hkl.blur_mtz(mtz, args.blur)
            mtz.write_to_file(mtz_in)
            st.cell = mtz.cell
            st.spacegroup_hm = mtz.spacegroup.hm
        elif args.hklin.endswith(".hkl"):
            asudata, hklf = utils.fileio.read_smcif_hkl(args.hklin, st.cell, sg_st)
            # TODO check consistency with model cell and sg
            write_mtz(mtz_in, asudata, hklf, args.blur)
            st.cell = asudata.unit_cell
            st.spacegroup_hm = asudata.spacegroup.hm
        else:
            raise SystemExit("Error: unsupported hkl file: {}".format(args.hklin))

    if args.invert:
        logger.writeln("Inversion of structure is requested.")
        old_sg = st.find_spacegroup()
        new_sg, tr = make_invert_tr(old_sg, st.cell)
        logger.writeln(" new space group = {} (no. {})".format(new_sg.hm, new_sg.number))
        st[0].transform_pos_and_adp(tr)
        if old_sg != new_sg:
            st.spacegroup_hm = new_sg.hm
            # overwrite mtz
            mtz = gemmi.read_mtz_file(mtz_in)
            mtz.spacegroup = new_sg
            mtz.write_to_file(mtz_in)

    if args.keyword_file:
        args.keyword_file = sum(args.keyword_file, [])
        for f in args.keyword_file:
            logger.writeln("Keyword file: {}".format(f))
            assert os.path.exists(f)
    else:
        args.keyword_file = []
            
    if args.keywords:
        args.keywords = sum(args.keywords, [])
    else:
        args.keywords = []


    # FIXME in some cases mtz space group should be modified. 
    utils.fileio.write_model(st, prefix="input", pdb=True, cif=True)

    if args.ligand: args.ligand = sum(args.ligand, [])

    prefix = "refined"
    if args.bref == "aniso":
        args.keywords.append("refi bref aniso")
    elif args.bref == "iso_then_aniso":
        prefix = "refined_1_iso"
        
    if args.unrestrained:
        args.keywords.append("refi type unre")
        args.no_hout = False
    else:
        monlib = utils.restraints.load_monomer_library(st, monomer_dir=args.monlib, cif_files=args.ligand, 
                                                       stop_for_unknowns=False)

    # no bulk solvent by default
    if not args.bulk_solvent:
        args.keywords.append("solvent no")

    # Run Refmac
    refmac = utils.refmac.Refmac(prefix=prefix, global_mode="cx",
                                 exe=args.exe,
                                 source=args.source,
                                 monlib_path=args.monlib,
                                 xyzin="input.mmcif",
                                 hklin=mtz_in,
                                 ncycle=args.ncycle,
                                 weight_matrix=args.weight_matrix,
                                 weight_auto_scale=args.weight_auto_scale,
                                 hydrogen=args.hydrogen,
                                 hout=not args.no_hout,
                                 resolution=args.resolution,
                                 keyword_files=args.keyword_file,
                                 keywords=args.keywords)
    refmac.set_libin(args.ligand)
    refmac_summary = refmac.run_refmac()

    if args.bref == "iso_then_aniso":
        refmac2 = refmac.copy(xyzin=prefix+".mmcif",
                              prefix="refined_2_aniso")
        refmac2.keywords.append("refi bref aniso")
        refmac_summary = refmac2.run_refmac()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)




"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.utils import restraints
from servalcat.utils import maps
import gemmi
import numpy
import pandas
import scipy.sparse
import os
import time
import itertools
import string

gemmi.IT92_normalize()
gemmi.Element("X").it92.set_coefs(gemmi.Element("O").it92.get_coefs()) # treat X (unknown) as O

u_to_b = 8 * numpy.pi**2
b_to_u = 1. / u_to_b

def shake_structure(st, sigma, copy=True):
    print("Randomizing structure with rmsd of {}".format(sigma))
    if copy:
        st2 = st.clone()
    else:
        st2 = st
        
    sigma /= numpy.sqrt(3)
    for model in st2:
        for cra in model.all():
            r = numpy.random.normal(0, sigma, 3)
            cra.atom.pos += gemmi.Position(*r)

    return st2
# shake_structure()

def setup_entities(st, clear=False, clear_entity_type=False, overwrite_entity_type=False, force_subchain_names=False):
    if clear: st.entities.clear()
    if clear_entity_type:
        for m in st:
            for c in m:
                for r in c:
                    r.entity_type = gemmi.EntityType.Unknown

    st.add_entity_types(overwrite_entity_type)
    st.assign_subchains(force_subchain_names)
    st.ensure_entities()
    st.deduplicate_entities()
# setup_entities()

def minimum_b(m):
    b_min = min((cra.atom.b_iso for cra in m.all()))
    eig_mins = [min(cra.atom.aniso.calculate_eigenvalues()) for cra in m.all() if cra.atom.aniso.nonzero()]
    if len(eig_mins) > 0: b_min = min(b_min, min(eig_mins) * u_to_b)
    return b_min
# minimum_b()

def determine_blur_for_dencalc(st, grid):
    b_min = minimum_b(st[0])
    b_need = grid**2*8*numpy.pi**2/1.1 # Refmac's way
    b_add = b_need - b_min
    return b_add
# determine_blur_for_dencalc()

def calc_sum_ab(st):
    sum_ab = dict()
    ret = 0.
    for cra in st[0].all():
        if cra.atom.element not in sum_ab:
            it92 = cra.atom.element.it92
            sum_ab[cra.atom.element] = sum(x*y for x,y in zip(it92.a, it92.b))
        ret += sum_ab[cra.atom.element] * cra.atom.occ
    return ret
# calc_sum_ab()

def calc_fc_fft(st, d_min, source, mott_bethe=True, monlib=None, blur=None, cutoff=1e-5, rate=1.5,
                omit_proton=False, omit_h_electron=False, miller_array=None):
    assert source in ("xray", "electron", "neutron")
    if source != "electron": assert not mott_bethe
    if omit_proton or omit_h_electron:
        assert mott_bethe
        if not st[0].has_hydrogen():
            logger.writeln("WARNING: omit_proton/h_electron requested, but no hydrogen exists!")
            omit_proton = omit_h_electron = False
        elif omit_proton and omit_h_electron:
            logger.writeln("omit_proton and omit_h_electron requested. removing hydrogens")
            st = st.clone()
            st.remove_hydrogens()
            omit_proton = omit_h_electron = False
    
    if blur is None: blur = determine_blur_for_dencalc(st, d_min/2/rate)
    #blur = max(0, blur) # negative blur may cause non-positive definite in case of anisotropic Bs
    logger.writeln("Setting blur= {:.2f} in density calculation (unblurred later)".format(blur))
        
    if mott_bethe and not omit_proton and monlib is not None and st[0].has_hydrogen():
        st = st.clone()
        topo = gemmi.prepare_topology(st, monlib, warnings=logger, ignore_unknown_links=True)
        resnames = st[0].get_all_residue_names()
        restraints.check_monlib_support_nucleus_distances(monlib, resnames)
        # Shift electron positions
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.ElectronCloud)
    else:
        topo = None
        
    if source == "xray" or mott_bethe:
        dc = gemmi.DensityCalculatorX()
    elif source == "electron":
        dc = gemmi.DensityCalculatorE()
    elif source == "neutron":
        dc = gemmi.DensityCalculatorN()
    else:
        raise RuntimeError("unknown source")

    dc.d_min = d_min
    dc.blur = blur
    dc.cutoff = cutoff
    dc.rate = rate
    dc.set_grid_cell_and_spacegroup(st)

    t_start = time.time()
    if mott_bethe:
        if omit_proton:
            method_str = "proton-omit Fc"
        elif omit_h_electron:
            if topo is None:
                method_str = "hydrogen electron-omit Fc"
            else:
                method_str = "hydrogen electron-omit, proton-shifted Fc"
        elif topo is not None:
            method_str = "proton-shifted Fc"
        else:
            method_str = "Fc"

        logger.writeln("Calculating {} using Mott-Bethe formula".format(method_str))
        
        dc.initialize_grid()
        dc.addends.subtract_z(except_hydrogen=True)

        if omit_h_electron:
            st2 = st.clone()
            st2.remove_hydrogens()
            dc.add_model_density_to_grid(st2[0])
        else:
            dc.add_model_density_to_grid(st[0])

        # Subtract hydrogen Z
        if not omit_proton and st[0].has_hydrogen():
            if topo is not None:
                # Shift proton positions
                topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus,
                                               default_scale=restraints.default_proton_scale)
            for cra in st[0].all():
                if cra.atom.is_hydrogen():
                    dc.add_c_contribution_to_grid(cra.atom, -1)

        dc.grid.symmetrize_sum()
        sum_ab = calc_sum_ab(st) * len(st.find_spacegroup().operations())
        mb_000 = sum_ab * gemmi.mott_bethe_const() / 4
    else:
        logger.writeln("Calculating Fc")
        dc.put_model_density_on_grid(st[0])
        mb_000 = 0

    logger.writeln(" done. Fc calculation time: {:.1f} s".format(time.time() - t_start))
    grid = gemmi.transform_map_to_f_phi(dc.grid)
    
    if miller_array is None:
        return grid.prepare_asu_data(dmin=d_min, mott_bethe=mott_bethe, unblur=dc.blur)
    else:
        return grid.get_value_by_hkl(miller_array, mott_bethe=mott_bethe, unblur=dc.blur,
                                     mott_bethe_000=mb_000)
# calc_fc_fft()

def calc_fc_direct(st, d_min, source, mott_bethe, monlib=None, miller_array=None):
    assert source in ("xray", "electron")
    if source != "electron": assert not mott_bethe

    miller_array_given = miller_array is not None
    unit_cell = st.cell
    spacegroup = gemmi.SpaceGroup(st.spacegroup_hm)
    if not miller_array_given: miller_array = gemmi.make_miller_array(unit_cell, spacegroup, d_min)
    topo = None

    if source == "xray" or mott_bethe:
        calc = gemmi.StructureFactorCalculatorX(st.cell)
    else:
        calc = gemmi.StructureFactorCalculatorE(st.cell)
        
    
    if source == "electron" and mott_bethe:
        if monlib is not None and st[0].has_hydrogen():
            st = st.clone()
            topo = gemmi.prepare_topology(st, monlib, warnings=logger, ignore_unknown_links=True)
            resnames = st[0].get_all_residue_names()
            restraints.check_monlib_support_nucleus_distances(monlib, resnames)

        calc.addends.clear()
        calc.addends.subtract_z(except_hydrogen=True)

    vals = []
    for hkl in miller_array:
        sf = calc.calculate_sf_from_model(st[0], hkl) # attention: traverse cell.images
        if mott_bethe: sf *= calc.mott_bethe_factor()
        vals.append(sf)

    if topo is not None:
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus,
                                       default_scale=restraints.default_proton_scale)

        for i, hkl in enumerate(miller_array):
            sf = calc.calculate_mb_z(st[0], hkl, only_h=True)
            if mott_bethe: sf *= calc.mott_bethe_factor()
            vals[i] += sf

    if miller_array_given:
        return numpy.array(vals)
    else:
        asu = gemmi.ComplexAsuData(unit_cell, spacegroup,
                                   miller_array, vals)
        return asu
# calc_fc_direct()

def get_em_expected_hydrogen(st, d_min, monlib, weights=None, blur=None, cutoff=1e-5, rate=1.5, optimize=False):
    # Very crude implementation to find peak from calculated map
    assert st[0].has_hydrogen()
    if blur is None: blur = determine_blur_for_dencalc(st, d_min/2/rate)
    blur = max(0, blur)
    logger.writeln("Setting blur= {:.2f} in density calculation".format(blur))

    st = st.clone()
    topo = gemmi.prepare_topology(st, monlib, warnings=logger, ignore_unknown_links=True)
    resnames = st[0].get_all_residue_names()
    restraints.check_monlib_support_nucleus_distances(monlib, resnames)

    topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.ElectronCloud)
    st_e = st.clone()
    topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus)
    st_n = st.clone()
    
    dc = gemmi.DensityCalculatorX()
    dc.d_min = d_min
    dc.blur = blur
    dc.cutoff = cutoff
    dc.rate = rate

    # Decide box_size
    max_r = max([dc.estimate_radius(cra.atom) for cra in st[0].all()])
    logger.writeln("max_r= {:.2f}".format(max_r))
    box_size = max_r*2 + 1 # padding
    logger.writeln("box_size= {:.2f}".format(box_size))
    mode_all = False #True
    if mode_all:
        dc.set_grid_cell_and_spacegroup(st)
    else:
        dc.grid.unit_cell = gemmi.UnitCell(box_size, box_size, box_size, 90, 90, 90)
        dc.grid.spacegroup = gemmi.SpaceGroup("P1")
        cbox = gemmi.Position(box_size/2, box_size/2, box_size/2)
    
    if mode_all: dc.initialize_grid()

    if weights is not None:
        w_s, w_w = weights # s_list and w_list
    else:
        w_s, w_w = None, None
        
    for ichain in range(len(st[0])):
        chain = st[0][ichain]
        for ires in range(len(chain)):
            residue = chain[ires]
            for iatom in range(len(residue)):
                atom = residue[iatom]
                if not atom.is_hydrogen(): continue
                h_n = st_n[0][ichain][ires][iatom]
                h_e = st_e[0][ichain][ires][iatom]
                if not mode_all:
                    dc.initialize_grid()
                    h_n.occ = 1.
                    h_e.occ = 1.
                    n_pos = gemmi.Position(h_n.pos)
                    h_n.pos = cbox
                    h_e.pos = cbox + h_e.pos - n_pos
                dc.add_atom_density_to_grid(h_e)
                dc.add_c_contribution_to_grid(h_n, -1)
                if not mode_all:
                    grid = gemmi.transform_map_to_f_phi(dc.grid)
                    asu_data = grid.prepare_asu_data(dmin=d_min, mott_bethe=True, unblur=dc.blur)
                    if w_s is not None:
                        asu_data.value_array[:] *= numpy.interp(1./asu_data.make_d_array(), w_s, w_w)
                        
                    denmap = asu_data.transform_f_phi_to_map(exact_size=(int(box_size*10), int(box_size*10), int(box_size*10)))
                    m = numpy.unravel_index(numpy.argmax(denmap), denmap.shape)
                    peakpos = denmap.get_position(m[0], m[1], m[2])
                    if optimize: peakpos = maps.optimize_peak(denmap, peakpos)
                    atom.pos = peakpos - cbox + n_pos

    if mode_all:
        grid = gemmi.transform_map_to_f_phi(dc.grid)
        asu_data = grid.prepare_asu_data(dmin=d_min, mott_bethe=True, unblur=dc.blur)
        if w_s is not None:
            asu_data.value_array[:] *= numpy.interp(1./asu_data.make_d_array(), w_s, w_w)
        denmap = asu_data.transform_f_phi_to_map(sample_rate=3)
        ccp4 = gemmi.Ccp4Map()
        ccp4.grid = denmap
        ccp4.update_ccp4_header(2, True) # float, update stats
        ccp4.write_ccp4_map("debug.ccp4")

    return st

# get_em_expected_hydrogen()

def translate_into_box(st, origin=None, apply_shift=True):
    if origin is None: origin = gemmi.Position(0,0,0)
    
    # apply unit cell translations to put model into a box (unit cell)
    omat = numpy.array(st.cell.orthogonalization_matrix)
    fmat = numpy.array(st.cell.fractionalization_matrix).transpose()
    com = numpy.array((st[0].calculate_center_of_mass() - origin).tolist())
    shift = sum([omat[:,i]*numpy.floor(1-numpy.dot(com, fmat[:,i])) for i in range(3)])
    tr = gemmi.Transform(gemmi.Mat33(), gemmi.Vec3(*shift))
    if apply_shift:
        for m in st: m.transform_pos_and_adp(tr)
    return shift
# translate_into_box()

def box_from_model(model, padding):
    allpos = numpy.array([cra.atom.pos.tolist() for cra in model.all()])
    ext = numpy.max(allpos, axis=0) - numpy.min(allpos, axis=0) + padding
    cell = gemmi.UnitCell(ext[0], ext[1], ext[2], 90, 90, 90)
    return cell
# box_from_model()

def cra_to_indices(cra, model):
    ret = [None, None, None]
    for ic in range(len(model)):
        chain = model[ic]
        if cra.chain != chain: continue
        ret[0] = ic
        for ir in range(len(chain)):
            res = chain[ir]
            if cra.residue != res: continue
            ret[1] = ir
            for ia in range(len(res)):
                if cra.atom == res[ia]:
                    ret[2] = ia

    return tuple(ret)
# cra_to_indices()

def cra_to_atomaddress(cra):
    aa = gemmi.AtomAddress(cra.chain.name,
                           cra.residue.seqid, cra.residue.name,
                           cra.atom.name, cra.atom.altloc)
    aa.res_id.segment = cra.residue.segment
    return aa
# cra_to_atomaddress()

def find_special_positions(st, special_pos_threshold=0.1, fix_occ=True, fix_pos=True, fix_adp=True):
    ns = gemmi.NeighborSearch(st[0], st.cell, 3).populate()
    cs = gemmi.ContactSearch(special_pos_threshold)
    cs.ignore = gemmi.ContactSearch.Ignore.SameAsu
    cs.special_pos_cutoff_sq = 0
    results = cs.find_contacts(ns)
    found = {}
    cra = {}
    for r in results:
        if r.partner1.atom != r.partner2.atom: continue
        found.setdefault(r.partner1.atom, []).append(r.image_idx)
        cra[r.partner1.atom] = r.partner1

    if found: logger.writeln("Atoms on special position detected.")
    ret = []
    for atom in found:
        images = found[atom]
        n_images = len(images) + 1
        sum_occ = atom.occ * n_images
        logger.writeln(" {} multiplicity= {} images= {} occupancies_total= {:.2f}".format(cra[atom], n_images, images, sum_occ))
        if sum_occ > 1.001 and fix_occ:
            new_occ = atom.occ / n_images
            logger.writeln("  correcting occupancy= {:.2f}".format(new_occ))
            atom.occ = new_occ
        if fix_pos:
            fpos = gemmi.Fractional(st.cell.frac.apply(atom.pos))
            fdiff = sum([(st.cell.images[i-1].apply(fpos) - fpos).wrap_to_zero() for i in images], gemmi.Fractional(0,0,0)) / n_images
            diff = st.cell.orth.apply(fdiff)
            atom.pos += gemmi.Position(diff)
            logger.writeln("  correcting position= {} (diff= {})".format(atom.pos.tolist(), diff.tolist()))
        if fix_adp and atom.aniso.nonzero():
            fani = atom.aniso.transformed_by(st.cell.frac.mat)
            fani_avg = sum([fani.transformed_by(st.cell.images[i-1].mat) for i in images], fani).scaled(1/n_images)
            atom.aniso = fani_avg.transformed_by(st.cell.orth.mat)
            logger.writeln("  correcting aniso= {}".format(atom.aniso.elements_pdb()))

        mat_total = (numpy.identity(3) + sum(numpy.array(st.cell.images[i-1].mat) for i in images)) / n_images
        ret.append((atom, images, mat_total))

    return ret
# find_special_positions()    

def expand_ncs(st, special_pos_threshold=0.01, howtoname=gemmi.HowToNameCopiedChain.Short):
    # TODO modify st.connections for atoms at special positions
    if len(st.ncs) == 0: return
    n_chains = len(st[0]) # number of chains before expansion
    specs = find_special_positions(st, special_pos_threshold)
    lookup = {x[0]:None for x in specs}
    for i, chain in enumerate(st[0]):
        for j, res in enumerate(chain):
            for k, atom in enumerate(res):
                for x in lookup:
                    if x == atom: lookup[x] = (i, j, k)
    logger.writeln("Expanding symmetry..")
    st.expand_ncs(howtoname)
    if not specs: return
    cs_count = len(st.find_spacegroup().operations())
    todel = []
    for atom, images, _ in specs:
        if not lookup[atom]: continue # should not happen
        ic, ir, ia = lookup[atom]
        mult = 1
        for img in images:
            # ignore crystallographic symmetry
            if (img - cs_count + 1) % cs_count != 0: continue
            mult += 1
            idx = img - cs_count + 1
            todel.append((idx * n_chains + ic, ir, ia))
        # correct occupancy
        st[0][ic][ir][ia].occ *= mult
    for ic, ir, ia in sorted(todel, reverse=True):
        del st[0][ic][ir][ia]
# expand_ncs()

def prepare_assembly(name, chains, ops, is_helical=False):
    a = gemmi.Assembly(name)
    g = gemmi.Assembly.Gen()
    if sum(map(lambda x: x.tr.is_identity(), ops)) == 0:
        g.operators.append(gemmi.Assembly.Operator()) # add identity
    for i, nop in enumerate(ops):
        op = gemmi.Assembly.Operator()
        op.transform = nop.tr
        if not nop.tr.is_identity():
            if is_helical:
                op.type = "helical symmetry operation"
            else:
                op.type = "point symmetry operation"
        g.operators.append(op)
    g.chains = chains
    a.generators.append(g)
    if is_helical:
        a.special_kind = gemmi.AssemblySpecialKind.RepresentativeHelical
    else:
        a.special_kind = gemmi.AssemblySpecialKind.CompletePoint
    return a
# prepare_assembly()

def filter_contacting_ncs(st, cutoff=5.):
    if len(st.ncs) == 0: return
    logger.writeln("Filtering out non-contacting NCS copies with cutoff={:.2f} A".format(cutoff))
    st.setup_cell_images()
    ns = gemmi.NeighborSearch(st[0], st.cell, cutoff*2).populate() # This is considered crystallographic cell if not 1 1 1. Undesirable result may be seen.
    cs = gemmi.ContactSearch(cutoff)
    cs.ignore = gemmi.ContactSearch.Ignore.SameAsu
    results = cs.find_contacts(ns)
    indices = set([r.image_idx for r in results])
    logger.writeln(" contacting copies: {}".format(indices))
    ops = [st.ncs[i-1] for i in indices] # XXX is this correct? maybe yes as long as identity operator is not there
    st.ncs.clear()
    st.ncs.extend(ops)
# filter_contacting_ncs()

def check_symmetry_related_model_duplication(st, distance_cutoff=0.5, max_allowed_ratio=0.5):
    logger.writeln("Checking if model in asu is given.")
    n_atoms = st[0].count_atom_sites()
    st.setup_cell_images()
    ns = gemmi.NeighborSearch(st[0], st.cell, 3).populate()
    cs = gemmi.ContactSearch(distance_cutoff)
    cs.ignore = gemmi.ContactSearch.Ignore.SameAsu
    results = cs.find_contacts(ns)
    n_contacting_atoms = len(set([a for r in results for a in (r.partner1.atom, r.partner2.atom)]))
    logger.writeln(" N_atoms= {} N_contacting_atoms= {}".format(n_atoms, n_contacting_atoms))
    return n_contacting_atoms / n_atoms > max_allowed_ratio # return True if too many contacts
# check_symmetry_related_model_duplication()

def adp_analysis(st, ignore_zero_occ=True):
    logger.writeln("= ADP analysis =")
    if ignore_zero_occ:
        logger.writeln("(zero-occupancy atoms are ignored)")

    all_B = []
    for i, mol in enumerate(st):
        if len(st) > 1: logger.writeln("Model {}:".format(i))
        logger.writeln("            min    Q1   med    Q3   max")
        stats = adp_stats_per_chain(mol, ignore_zero_occ)
        for chain, natoms, qs in stats:
            logger.writeln(("Chain {:3s}".format(chain) if chain!="*" else "All      ") + " {:5.1f} {:5.1f} {:5.1f} {:5.1f} {:5.1f}".format(*qs))
    logger.writeln("")
# adp_analysis()

def adp_stats_per_chain(model, ignore_zero_occ=True):
    bs = {}
    for cra in model.all():
        if not ignore_zero_occ or cra.atom.occ > 0:
            bs.setdefault(cra.chain.name, []).append(cra.atom.b_iso)

    ret = []
    for chain in model:
        if chain.name in [x[0] for x in ret]: continue
        qs = numpy.quantile(bs[chain.name], [0,0.25,0.5,0.75,1])
        ret.append((chain.name, len(bs[chain.name]), qs))

    if len(bs) > 1:
        all_bs = sum(bs.values(), [])
        qs = numpy.quantile(all_bs, [0,0.25,0.5,0.75,1])
        ret.append(("*", len(all_bs), qs))
        
    return ret
# adp_stats_per_chain()

def reset_adp(model, bfactor=None, is_aniso=False):
    for cra in model.all():
        if bfactor is not None:
            cra.atom.b_iso = bfactor
        if not is_aniso:
            cra.atom.aniso = gemmi.SMat33f(0,0,0,0,0,0)
        else:
            if not cra.atom.aniso.nonzero() or bfactor is not None:
                u = cra.atom.b_iso * b_to_u
                cra.atom.aniso = gemmi.SMat33f(u, u, u, 0, 0, 0)
# reset_adp()

def shift_b(model, b):
    u = b * b_to_u
    for cra in model.all():
        cra.atom.b_iso += b
        if cra.atom.aniso.nonzero():
            cra.atom.aniso.u11 += u
            cra.atom.aniso.u22 += u
            cra.atom.aniso.u33 += u
# shift_b()

def all_chain_ids(st):
    return [chain.name for model in st for chain in model]
# all_chain_ids()

def all_B(st, ignore_zero_occ=True):
    ret = []
    for mol in st:
        for cra in mol.all():
            if not ignore_zero_occ or cra.atom.occ > 0:
                ret.append(cra.atom.b_iso)

    return ret
# all_B()

def to_dataframe(st):
    keys = ("model", "chain", "resn", "subchain", "segment", "seqnum", "icode", "altloc",
            "u11", "u22", "u33", "u12", "u13", "u23",
            "b_iso", "charge", "elem", "atom", "occ",
            "x", "y", "z", "tlsgroup")
    d = dict([(x,[]) for x in keys])
    app = lambda k, v: d[k].append(v)
    
    for m in st:
        for cra in m.all():
            c,r,a = cra.chain, cra.residue, cra.atom
            # TODO need support r.het_flag, r.flag, a.calc_flag, a.flag, a.serial?
            app("model", m.name)
            app("chain", c.name)
            app("resn", r.name)
            app("subchain", r.subchain)
            app("segment", r.segment)
            app("seqnum", r.seqid.num)
            app("icode", r.seqid.icode)
            app("altloc", a.altloc)
            app("u11", a.aniso.u11)
            app("u22", a.aniso.u22)
            app("u33", a.aniso.u33)
            app("u12", a.aniso.u12)
            app("u13", a.aniso.u13)
            app("u23", a.aniso.u23)
            app("b_iso", a.b_iso)
            app("charge", a.charge)
            app("elem", a.element.name)
            app("atom", a.name)
            app("occ", a.occ)
            app("x", a.pos.x)
            app("y", a.pos.y)
            app("z", a.pos.z)
            app("tlsgroup", a.tls_group_id)

    return pandas.DataFrame(data=d)
# to_dataframe()

def from_dataframe(df, st=None): # Slow!
    if st is None:
        st = gemmi.Structure()
    else:
        st = st.clone()
        for i in range(len(st)):
            del st[0]
        
    for m_name, dm in df.groupby("model"):
        st.add_model(gemmi.Model(m_name))
        m = st[-1]
        for c_name, dc in dm.groupby("chain"):
            m.add_chain(gemmi.Chain(c_name))
            c = m[-1]
            for rkey, dr in dc.groupby(["seqnum","icode","resn","segment","subchain"]):
                c.add_residue(gemmi.Residue())
                r = c[-1]
                r.seqid.num = rkey[0]
                r.seqid.icode = rkey[1]
                r.name = rkey[2]
                r.segment = rkey[3]
                r.subchain = rkey[4]
                for _, row in dr.iterrows():
                    r.add_atom(gemmi.Atom())
                    a = r[-1]
                    a.altloc = row["altloc"]
                    a.name = row["atom"]
                    a.aniso.u11 = row["u11"]
                    a.aniso.u22 = row["u22"]
                    a.aniso.u33 = row["u33"]
                    a.aniso.u12 = row["u12"]
                    a.aniso.u13 = row["u13"]
                    a.aniso.u23 = row["u23"]
                    a.b_iso = row["b_iso"]
                    a.charge = row["charge"]
                    a.element = gemmi.Element(row["elem"])
                    a.occ = row["occ"]
                    a.pos.x = row["x"]
                    a.pos.y = row["y"]
                    a.pos.z = row["z"]
                    a.tls_group_id = row["tlsgroup"]
                    
    return st
# from_dataframe()

def st_from_positions(positions, bs=None, qs=None):
    st = gemmi.Structure()
    st.add_model(gemmi.Model("1"))
    st[0].add_chain(gemmi.Chain("A"))
    c = st[0][0]
    if bs is None: bs = (0. for _ in range(len(positions)))
    if qs is None: qs = (1. for _ in range(len(positions)))
    for i, (pos, b, q) in enumerate(zip(positions, bs, qs)):
        c.add_residue(gemmi.Residue())
        r = c[-1]
        r.seqid.num = i
        r.name = "HOH"
        r.add_atom(gemmi.Atom())
        a = r[-1]
        a.name = "O"
        a.element = gemmi.Element("O")
        a.pos = pos
        a.b_iso = b
        a.occ = q
                    
    return st
# st_from_positions()
            
def invert_model(st):
    # invert x-axis
    A = numpy.array(st.cell.orthogonalization_matrix.tolist())
    center = numpy.sum(A,axis=1) / 2
    center = gemmi.Vec3(*center)
    mat = gemmi.Mat33([[-1,0,0],[0,1,0],[0,0,1]]) 
    vec = mat.multiply(-center) + center
    tr = gemmi.Transform(mat, vec)
    st[0].transform_pos_and_adp(tr)

    # invert peptides
# invert_model()

def cx_to_mx(ss): #SmallStructure to Structure
    st = gemmi.Structure()
    st.spacegroup_hm = ss.spacegroup_hm
    st.cell = ss.cell
    st.add_model(gemmi.Model("1"))
    st[-1].add_chain(gemmi.Chain("A"))
    st[-1][-1].add_residue(gemmi.Residue())
    st[-1][-1][-1].seqid.num = 1
    st[-1][-1][-1].name = "00"

    ruc = ss.cell.reciprocal()
    cif2cart = ss.cell.orthogonalization_matrix.multiply_by_diagonal(gemmi.Vec3(ruc.a, ruc.b, ruc.c))
    as_smat33f = lambda x: gemmi.SMat33f(x.u11, x.u22, x.u33, x.u12, x.u13, x.u23)
    
    for site in ss.sites:
        st[-1][-1][-1].add_atom(gemmi.Atom())
        a = st[-1][-1][-1][-1]
        a.name = site.label
        a.aniso = as_smat33f(site.aniso.transformed_by(cif2cart))
        a.b_iso = site.u_iso * u_to_b
        #a.charge = ?
        a.element = site.element
        a.occ = site.occ
        a.pos = site.orth(ss.cell)
        
    return st
# cx_to_mx()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import json
import os
import shutil
import argparse
from servalcat.utils import logger
from servalcat import utils
from servalcat.xtal.sigmaa import process_input, calculate_maps
from servalcat.refine.xtal import LL_Xtal
from servalcat.refine.refine import Geom, Refine
b_to_u = utils.model.b_to_u

def add_arguments(parser):
    parser.description = "EXPERIMENTAL program to refine crystallographic structures"
    parser.add_argument("--hklin", required=True)
    parser.add_argument("-d", '--d_min', type=float)
    parser.add_argument('--nbins', type=int, default=20,
                        help="Number of bins (default: %(default)d)")
    parser.add_argument("--labin", nargs="+", help="F SIGF FREE input")
    parser.add_argument('--free', type=int, default=0,
                        help='flag number for test set')
    parser.add_argument('--model', required=True,
                        help='Input atomic model file')
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('--ligand', nargs="*", action="append",
                        help="restraint dictionary cif file(s)")
    parser.add_argument('--hydrogen', default="all", choices=["all", "yes", "no"],
                        help="all: add riding hydrogen atoms, yes: use hydrogen atoms if present, no: remove hydrogen atoms in input. "
                        "Default: %(default)s")
    parser.add_argument('--jellybody', action='store_true',
                        help="Use jelly body restraints")
    parser.add_argument('--jellybody_params', nargs=2, type=float,
                        metavar=("sigma", "dmax"), default=[0.01, 4.2],
                        help="Jelly body sigma and dmax (default: %(default)s)")
    parser.add_argument('--keywords', nargs='+', action="append",
                        help="refmac keyword(s)")
    parser.add_argument('--keyword_file', nargs='+', action="append",
                        help="refmac keyword file(s)")
    parser.add_argument('--randomize', type=float, default=0,
                        help='Shake coordinates with specified rmsd')
    parser.add_argument('--ncycle', type=int, default=10,
                        help="number of CG cycles (default: %(default)d)")
    parser.add_argument('--weight', type=float, default=1,
                        help="refinement weight")
    parser.add_argument('--sigma_b', type=float, default=30,
                        help="refinement ADP sigma in B (default: %(default)f)")
    parser.add_argument('--bfactor', type=float,
                        help="reset all atomic B values to specified value")
    parser.add_argument('--fix_xyz', action="store_true")
    parser.add_argument('--adp',  choices=["fix", "iso", "aniso"], default="iso")
    parser.add_argument('--max_dist_for_adp_restraint', type=float, default=4.)
    parser.add_argument('--unrestrained',  action='store_true', help="No positional restraints")
    parser.add_argument('--refine_h', action="store_true", help="Refine hydrogen (default: restraints only)")
    parser.add_argument("--source", choices=["electron", "xray", "neutron"], default="electron")
    parser.add_argument('--no_solvent',  action='store_true',
                        help="Do not consider bulk solvent contribution")
    parser.add_argument('-o','--output_prefix')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def main(args):
    if not args.output_prefix:
        args.output_prefix = utils.fileio.splitext(os.path.basename(args.model))[0] + "_refined"

    keywords = []
    if args.keywords or args.keyword_file:
        if args.keywords: keywords = sum(args.keywords, [])
        if args.keyword_file: keywords.extend(l for f in sum(args.keyword_file, []) for l in open(f))

    hkldata, sts, fc_labs, centric_and_selections = process_input(hklin=args.hklin,
                                                                  labin=args.labin,
                                                                  n_bins=args.nbins,
                                                                  free=args.free,
                                                                  xyzins=[args.model],
                                                                  source=args.source,
                                                                  d_min=args.d_min)
    st = sts[0]
    monlib = utils.restraints.load_monomer_library(st, monomer_dir=args.monlib, cif_files=args.ligand,
                                                   stop_for_unknowns=False)
    h_change = {"all":gemmi.HydrogenChange.ReAddButWater,
                "yes":gemmi.HydrogenChange.NoChange,
                "no":gemmi.HydrogenChange.Remove}[args.hydrogen]
    try:
        topo = utils.restraints.prepare_topology(st, monlib, h_change=h_change,
                                                 check_hydrogen=(args.hydrogen=="yes"))
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    # initialize ADP
    if args.adp != "fix":
        utils.model.reset_adp(st[0], args.bfactor, args.adp == "aniso")
    
    geom = Geom(st, topo, monlib, shake_rms=args.randomize, sigma_b=args.sigma_b, refmac_keywords=keywords)
    geom.geom.adpr_max_dist = args.max_dist_for_adp_restraint
    if args.jellybody:
        geom.geom.ridge_sigma, geom.geom.ridge_dmax = args.jellybody_params

    ll = LL_Xtal(hkldata, centric_and_selections, args.free, st, monlib, source=args.source, use_solvent=not args.no_solvent)
    refiner = Refine(st, geom, ll=ll,
                     refine_xyz=not args.fix_xyz,
                     adp_mode=dict(fix=0, iso=1, aniso=2)[args.adp],
                     refine_h=args.refine_h,
                     unrestrained=args.unrestrained)

    refiner.run_cycles(args.ncycle, weight=args.weight)
    utils.fileio.write_model(refiner.st, args.output_prefix, pdb=True, cif=True)

    calculate_maps(ll.hkldata, centric_and_selections, ll.fc_labs, ll.D_labs, args.output_prefix + "_stats.log")

    # Write mtz file
    labs = ["FP", "SIGFP", "FOM", "FWT", "DELFWT", "FC"]
    if not args.no_solvent:
        labs.append("FCbulk")
    mtz_out = args.output_prefix+".mtz"
    hkldata.write_mtz(mtz_out, labs=labs, types={"FOM": "W", "FP":"F", "SIGFP":"Q"})
    logger.writeln("output mtz: {}".format(mtz_out))

# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
from servalcat.utils import logger
from servalcat.refmac.refmac_keywords import get_lines, read_exte_line

"""import line_profiler
profile = line_profiler.LineProfiler()
import atexit
atexit.register(profile.print_stats)
@profile"""
def read_external_restraints(lines, st, geom):
    # default or current values
    defs = dict(symall_block=False, exclude_self_block=False, type_default=2, alpha_default=1.,
                ext_verbose=False, scale_sigma_dist=1., scale_sigma_angl=1., scale_sigma_tors=1.,
                scale_sigma_chir=1., scale_sigma_plan=1., scale_sigma_inte=1.,
                sigma_min_loc=0., sigma_max_loc=100., ignore_undefined=False, ignore_hydrogens=True,
                dist_max_external=numpy.inf, dist_min_external=-numpy.inf, use_atoms="a", prefix_ch=" ")
    #exte = gemmi.ExternalRestraints(st)
    extypes = dict(dist=gemmi.Geometry.Bond,
                   angl=gemmi.Geometry.Angle,
                   chir=gemmi.Geometry.Chirality,
                   tors=gemmi.Geometry.Torsion,
                   plan=gemmi.Geometry.Plane,
                   inte=gemmi.Geometry.Interval,
                   harm=gemmi.Geometry.Harmonic,
                   spec=gemmi.Geometry.Special,
                   stac=gemmi.Geometry.Stacking)
    exlists = dict(dist=geom.bonds, angl=geom.angles, tors=geom.torsions,
                   chir=geom.chirs, plan=geom.planes, inte=geom.intervals,
                   stac=geom.stackings, harm=geom.harmonics, spec=geom.specials)
    num_org = {x: len(exlists[x]) for x in exlists}

    # XXX There may be duplication (same chain, resi, name, and alt) - we should give error?
    lookup = {(cra.chain.name, cra.residue.seqid.num, cra.residue.seqid.icode,
               cra.atom.name, cra.atom.altloc) : cra.atom for cra in st[0].all()}

    # TODO main chain / side chain filtering, hydrogen, dist_max_external/dist_min_external
    for l in get_lines(lines):
        r = read_exte_line(l)
        if not r: continue
        defs.update(r["defaults"])
        if "rest_type" not in r: continue
        if r["rest_type"] not in extypes:
            logger.writeln("Warning: unknown external restraint type: {}".format(r["rest_type"]))
            continue

        atoms = []
        skip = False
        for i, spec in enumerate(r["restr"].get("specs", [])):
            if r["rest_type"] == "stac":
                atoms.append([])
            if "ifirst" in spec:
                for chain in st[0]:
                    if chain.name != spec["chain"]: continue
                    for res in chain:
                        if spec["ifirst"] is not None and res.seqid.num < spec["ifirst"]: continue
                        if spec["ilast"] is not None and res.seqid.num > spec["ilast"]: continue
                        atoms.extend([a for a in res if spec["atom"] == "*" or a.name == spec["atom"]])
            else:
                for name in spec["names"]: # only same altloc allowed?
                    key = (spec["chain"], spec["resi"], spec.get("icode", " "),
                           name, spec.get("altloc", "\0"))
                    atom = lookup.get(key)
                    if atom is None:
                        if defs["ignore_undefined"]:
                            logger.writeln("Warning: atom not found: {}\n=>{}".format(key, l))
                            skip = True
                            continue
                        raise RuntimeError("Atom not found: {}\n=>{}".format(key, l))
                    if r["rest_type"] == "stac":
                        atoms[i].append(atom)
                    else:
                        atoms.append(atom)
        if skip or not atoms:
            continue
        if r["rest_type"] in ("spec", "harm"):
            if r["restr"]["rectype"] == "auto":
                assert r["rest_type"] == "spec"
                atoms = [cra.atom for cra in st[0].all()]
            for atom in atoms:
                ex = extypes[r["rest_type"]](atom)
                if r["rest_type"] == "spec":
                    # TODO check if it is on special position. using r["restr"]["toler"]
                    ex.sigma_t = r["restr"]["sigma_t"]
                    ex.sigma_u =r["restr"]["sigma_u"]
                    ex.u_val_incl = r["restr"]["u_val_incl"]
                    # ex.trans_t =
                    # ex.mat_u = 
                else:
                    ex.sigma = r["restr"]["sigma_t"]
                exlists[r["rest_type"]].append(ex)
            continue
        elif r["rest_type"] == "plan":
            ex = extypes[r["rest_type"]](atoms)
        else:
            ex = extypes[r["rest_type"]](*atoms)
        if r["rest_type"] in ("dist", "angl", "chir", "tors"):
            value = r["restr"]["value"]
            sigma = r["restr"]["sigma_value"] / defs["scale_sigma_{}".format(r["rest_type"])]
            if r["rest_type"] == "chir":
                ex.value = value
                ex.sigma = sigma
            else:
                if r["rest_type"] == "dist":
                    sigma = min(max(sigma, defs["sigma_min_loc"]), defs["sigma_max_loc"])
                    vals = (value, sigma, value, sigma) # nucleus
                elif r["rest_type"] == "tors":
                    vals = (value, sigma, 1) # period. # Refmac does not seem to read it from instruction
                else:
                    vals = (value, sigma)
                ex.values.append(extypes[r["rest_type"]].Value(*vals))
        
        if r["rest_type"] == "dist":
            if not (defs["dist_min_external"] < r["restr"]["value"] < defs["dist_max_external"]):
                continue
            ex.alpha = r["restr"].get("alpha_in", defs["alpha_default"])
            ex.type = r["restr"].get("itype_in", defs["type_default"])
            symm1 = any([spec.get("symm") for spec in r["restr"]["specs"]]) # is it the intention?
            if r["restr"].get("symm_in", defs["symall_block"]) or symm1:
                asu = gemmi.Asu.Different if defs["exclude_self_block"] else gemmi.Asu.Any
                im = st.cell.find_nearest_image(ex.atoms[0].pos, ex.atoms[1].pos, asu)
                ex.set_image(im)
            print("dist=", ex.alpha, ex.type, ex.values[-1].value, ex.values[-1].sigma, ex.sym_idx, ex.pbc_shift, ex.atoms)
        elif r["rest_type"] == "angl":
            print("angl=", ex.values[-1].value, ex.values[-1].sigma, ex.atoms)
        elif r["rest_type"] == "tors":
            print("tors=", ex.values[-1].value, ex.values[-1].sigma, ex.atoms)
        elif r["rest_type"] == "chir":
            print("chir=", ex.value, ex.sigma, ex.atoms)
            ex.sign = gemmi.ChiralityType.Positive if ex.value > 0 else gemmi.ChiralityType.Negative
            ex.value = abs(ex.value)
        elif r["rest_type"] == "plan":
            ex.sigma = r["restr"]["sigma_value"] / defs["scale_sigma_{}".format(r["rest_type"])]
            print("plan=", ex.sigma, ex.atoms)
        elif r["rest_type"] == "inte":
            dmin, dmax = r["restr"].get("dmin"), r["restr"].get("dmax")
            smin, smax = r["restr"].get("smin"), r["restr"].get("smax")
            if (smin,smax).count(None) == 2:
                smin = smax = 0.05
            else:
                if smin is None: smin = smax
                if smax is None: smax = smin
                smin /= defs["scale_sigma_inte"]
                smax /= defs["scale_sigma_inte"]
            if (dmin,dmax).count(None) == 1:
                if dmin is None: dmin = dmax
                if dmax is None: dmax = dmin
            ex.dmin = dmin
            ex.dmax = dmax
            ex.smin = smin
            ex.smax = smax
            print("inte=", ex.dmin, ex.dmax, ex.smin, ex.smax, ex.atoms)
        elif r["rest_type"] == "stac":
            ex.dist = r["restr"]["dist_id"]
            ex.sd_dist = r["restr"]["dist_sd"]
            ex.angle = r["restr"].get("angle_id", 0.)
            ex.sd_angle = r["restr"]["angle_sd"]
            print("stac=", ex.dist, ex.sd_dist, ex.angle, ex.sd_angle, ex.planes)
            
        exlists[r["rest_type"]].append(ex)

    logger.writeln("External restraints from Refmac instructions")
    labs = dict(dist="distances", angl="angles", tors="torsions",
                chir="chirals", plan="planes", inte="intervals",
                stac="stackings", harm="harmonics", spec="special positions")
    for lab in labs:
        logger.writeln(" Number of {:18s} : {}".format(labs[lab], len(exlists[lab]) - num_org[lab]))
    logger.writeln("")
# read_external_restraints()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import argparse
import sys
import traceback
import platform
import gemmi
import numpy
import scipy
import pandas
import servalcat.spa.shiftback
import servalcat.spa.run_refmac
import servalcat.spa.fsc
import servalcat.spa.fofc
import servalcat.spa.shift_maps
import servalcat.spa.translate
import servalcat.spa.localcc
import servalcat.xtal.run_refmac_small
import servalcat.xtal.sigmaa
import servalcat.refmac.refmac_wrapper
import servalcat.utils.commands
import servalcat.refine.refine_geom
import servalcat.refine.refine_spa
import servalcat.refine.refine_xtal

from servalcat.utils import logger

def test_installation():
    vers = logger.dependency_versions()
    pandas_ver = [int(x) for x in vers["pandas"].split(".")]
    numpy_ver = [int(x) for x in vers["numpy"].split(".")]
    msg_unknown = "Unexpected error occurred (related to numpy+pandas). Please report to authors with the result of servalcat -v."
    msg_skip = "If you want to ignore this error, please specify --skip_test."
    ret = True
    
    try:
        x = pandas.DataFrame(dict(x=[2j]))
        x.merge(x)
    except TypeError:
        ret = False
        if pandas_ver >= [1,3,0] and numpy_ver < [1,19,1]:
            print("There is a problem in pandas+numpy. Please update numpy to 1.19.1 or newer (or use pandas < 1.3.0).")
        else:
            print(traceback.format_exc())
            print(msg_unknown)
    except:
        print(traceback.format_exc())
        print(msg_unknown)
        ret = False

    if not ret:
        print(msg_skip)
        
    return ret
# test_installation()        

def main():
    parser = argparse.ArgumentParser(prog="servalcat",
                                     description="A tool for model refinement and map calculation for cryo-EM SPA.")
    parser.add_argument("--skip_test", action="store_true", help="Skip installation test")
    parser.add_argument("-v", "--version", action="version",
                        version="Servalcat {servalcat} with Python {python} ({deps})".format(servalcat=servalcat.__version__,
                                                                                             python=platform.python_version(),
                                                                                             deps=", ".join([x[0]+" "+x[1] for x in logger.dependency_versions().items()])))
    subparsers = parser.add_subparsers(dest="command")

    modules = dict(shiftback=servalcat.spa.shiftback,
                   refine_spa=servalcat.spa.run_refmac,
                   refine_cx=servalcat.xtal.run_refmac_small,
                   fsc=servalcat.spa.fsc,
                   fofc=servalcat.spa.fofc,
                   trim=servalcat.spa.shift_maps,
                   translate=servalcat.spa.translate,
                   localcc=servalcat.spa.localcc,
                   sigmaa=servalcat.xtal.sigmaa,
                   #show=servalcat.utils.show,
                   util=servalcat.utils.commands,
                   refmac5=servalcat.refmac.refmac_wrapper,
                   refine_geom=servalcat.refine.refine_geom,
                   refine_spa_norefmac=servalcat.refine.refine_spa,
                   refine_xtal_norefmac=servalcat.refine.refine_xtal,
                   )

    for n in modules:
        p = subparsers.add_parser(n)
        modules[n].add_arguments(p)

    args = parser.parse_args()
    
    if not args.skip_test and not test_installation():
        return
    
    if args.command == "util" and not args.subcommand:
        print("specify subcommand.")    
    elif args.command in modules:
        logger.set_file("servalcat.log")
        logger.write_header()
        try:
            modules[args.command].main(args)
        except SystemExit as e:
            logger.error(str(e))
            sys.exit(1)
    else:
        parser.print_help()

# main()

if __name__ == "__main__":
    main()




"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import pandas
from servalcat.utils import logger
from servalcat import utils
from servalcat import spa

def add_arguments(parser):
    parser.description = 'Find translation of the model in the map'
    parser.add_argument('--model',
                        required=True,
                        help="")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--halfmaps", nargs=2, help="Input half map files")
    group.add_argument("--map", help="Use this only if you really do not have half maps.")
    parser.add_argument('--mask',
                        help='Mask file')
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    parser.add_argument('-d', '--resolution',
                        type=float,
                        required=True,
                        help='')
    parser.add_argument('--no_interpolation', action="store_true",
                        help="No interpolation in peak finding of translation function")
    parser.add_argument('-o', '--output_prefix', default="translated")

# add_arguments()
                        
def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def calc_fsc(hkldata, lab1, lab2):
    stats = hkldata.binned_df[["d_min", "d_max"]].copy()
    stats["ncoeffs"] = 0
    stats["fsc"] = 0.
    for i_bin, idxes in hkldata.binned():
        stats.loc[i_bin, "ncoeffs"] = len(idxes)
        stats.loc[i_bin, "fsc"] = numpy.real(numpy.corrcoef(hkldata.df[lab1].to_numpy()[idxes],
                                                            hkldata.df[lab2].to_numpy()[idxes])[1,0])

    sum_n = sum(stats.ncoeffs)
    fscavg = sum(stats.ncoeffs*stats.fsc)/sum_n
    return stats, fscavg
# calc_fsc()

def find_peak(tf_map, ini_pos):
    logger.writeln("Finding peak using interpolation..")

    x = tf_map.unit_cell.fractionalize(ini_pos)
    logger.writeln("       x0: [{}, {}, {}]".format(*x.tolist()))
    logger.writeln("       f0: {}".format(-tf_map.tricubic_interpolation(x)))
    
    res = scipy.optimize.minimize(fun=lambda x:-tf_map.tricubic_interpolation(gemmi.Fractional(*x)),
                                  x0=x.tolist(),
                                  jac=lambda x:-numpy.array(tf_map.tricubic_interpolation_der(gemmi.Fractional(*x))[1:]))
    logger.writeln(str(res))
    final_pos = tf_map.unit_cell.orthogonalize(gemmi.Fractional(*res.x))
    logger.writeln(" Move from initial: [{:.3f}, {:.3f}, {:.3f}] A".format(*(final_pos-ini_pos).tolist()))
    return final_pos
# find_peak()

def main(args):
    if args.halfmaps:
        maps = utils.fileio.read_halfmaps(args.halfmaps, pixel_size=args.pixel_size)
        assert maps[0][0].shape == maps[1][0].shape
        assert maps[0][0].unit_cell == maps[1][0].unit_cell
        assert maps[0][1] == maps[1][1]
    else:
        maps = [utils.fileio.read_ccp4_map(args.map, pixel_size=args.pixel_size)]

    model_format = utils.fileio.check_model_format(args.model)
    st = utils.fileio.read_structure(args.model)
    st.cell = maps[0][0].unit_cell
    st.spacegroup_hm = "P1"

    if args.mask:
        mask = utils.fileio.read_ccp4_map(args.mask)[0]
    else:
        mask = None

    hkldata = utils.maps.mask_and_fft_maps(maps, args.resolution, mask=None)
    hkldata.df["FC"] = utils.model.calc_fc_fft(st, args.resolution - 1e-6, source="electron",
                                               miller_array=hkldata.miller_array())
    hkldata.setup_relion_binning()

    stats, fscavg = calc_fsc(hkldata, "FP", "FC")
    logger.writeln(stats.to_string())
    logger.writeln("FSCaverage before translation = {:.4f}".format(fscavg))

    hkldata.df["TF"] = hkldata.df.FP.to_numpy() * numpy.conj(hkldata.df.FC.to_numpy())

    tf_map = hkldata.fft_map("TF")
    max_idx = numpy.unravel_index(numpy.argmax(tf_map), tf_map.shape)
    shift = tf_map.get_position(*max_idx)

    if not args.no_interpolation:
        shift = utils.maps.optimize_peak(tf_map, shift)

    logger.writeln("shift= {:.4f}, {:.4f}, {:.4f} ".format(*shift))

    # phase shift for translation
    hkldata.df.FC *= numpy.exp(2.j*numpy.pi*numpy.dot(hkldata.miller_array(),
                                                      hkldata.cell.fractionalize(shift).tolist()))
    stats, fscavg = calc_fsc(hkldata, "FP", "FC")
    logger.writeln(stats.to_string())
    logger.writeln("FSCaverage after translation = {:.4f}".format(fscavg))

    tr = gemmi.Transform(gemmi.Mat33(), shift)
    st[0].transform_pos_and_adp(tr)
    utils.model.translate_into_box(st)
    utils.fileio.write_model(st, file_name=args.output_prefix+model_format)
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)
